import os
import pickle
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from matplotlib.ticker import FormatStrFormatter


class GridSearchContainer:
    runparams: dict
    run_ids_clean: dict
    evals: dict
    evals_posterior_sampling: dict
    histories: dict
    source_gs: dict
    cv_ids: dict

    target_cell_runparams: dict
    target_cell_evals: dict
    target_cell_indices: dict

    keys_all: list
    keys_test: list
    keys_train: list
    keys_val: list
    nodes_test: Dict[str, np.ndarray]
    nodes_train: Dict[str, np.ndarray]
    nodes_val: Dict[str, np.ndarray]

    def __init__(self, source_path, gs_ids, lateral_resolution):
        self.source_path = source_path
        if isinstance(gs_ids, str):
            gs_ids = [gs_ids]
        self.gs_ids = gs_ids
        self.lateral_resolution = lateral_resolution

        self.summary_table = None
        self.runparams_table = None
        self.target_cell_table = None
        self.best_model_hyperparam_table = None
        self.target_cell_frequencies = None

    @property
    def cv_keys(self) -> List[str]:
        """
        Returns keys of cross-validation used in dictionaries in this class.

        :return: list of string keys
        """
        return np.unique(self.summary_table["cv"].values).tolist()

    def load_gs(
        self,
        expected_pickle: Optional[list] = None,
        add_posterior_sampling_model: bool = False,
        report_unsuccessful_runs: bool = False,
    ):
        """
        Load all metrics from grid search output files.

        Core results are save in self.summary_table.

        :return:
        """
        if expected_pickle is None:
            expected_pickle = ["evaluation", "history", "hyperparam", "model_args", "time"]

        if add_posterior_sampling_model:
            expected_pickle += ["evaluation_posterior_sampling"]
        self.summary_table = []
        self.runparams_table = []
        self.runparams = {}
        self.run_ids_clean = {}

        self.evals = {}
        self.evals_posterior_sampling = {}
        self.histories = {}
        self.source_gs = {}
        self.cv_ids = {}
        for gs_id in self.gs_ids:
            # Collect runs that belong to grid search by looping over file names in directory.
            indir = self.source_path + gs_id + "/results/"
            # runs_ids are the unique hyper-parameter settings, which are again subsetted by cross-validation.
            # These ids are present in all files names but are only collected from the *model.tf file names here.

            run_ids = np.sort(
                np.unique(
                    [
                        "_".join(".".join(x.split(".")[:-1]).split("_")[:-1])
                        for x in os.listdir(indir)
                        if x.split("_")[-1].split(".")[0] == "time"
                    ]
                )
            )
            cv_ids = np.sort(np.unique([x.split("_")[-1] for x in run_ids]))  # identifiers of cross-validation splits
            run_ids = np.sort(
                np.unique(["_".join(x.split("_")[:-1]) for x in run_ids])  # identifiers of hyper-parameters settings
            )
            run_ids_clean = []  # only IDs of completed runs (all files present)
            for j, r in enumerate(run_ids):
                complete_run = True
                for cv in cv_ids:
                    # Check pickled files:
                    for end in expected_pickle:
                        fn = r + "_" + cv + "_" + end + ".pickle"
                        if not os.path.isfile(indir + fn):
                            if report_unsuccessful_runs:
                                print("File %r missing" % fn)
                            complete_run = False
                # Check run parameter files (one per cross-validation set):
                fn = r + "_runparams.pickle"
                if not os.path.isfile(indir + fn):
                    if report_unsuccessful_runs:
                        print("File %r missing" % fn)
                    complete_run = False
                if not complete_run:
                    if report_unsuccessful_runs:
                        print("Run %r not successful" % r + "_" + cv)
                else:
                    run_ids_clean.append(r)
            # Load results and settings from completed runs:
            evals = (
                {}
            )  # Dictionary over runs with dictionary over cross-validations with results from model evaluation.
            runparams = {}  # Dictionary over runs with model settings.
            histories = {}
            for x in run_ids_clean:
                # Load model settings (these are shared across all partitions).
                fn_runparams = indir + x + "_runparams.pickle"
                with open(fn_runparams, "rb") as f:
                    runparams[x] = pickle.load(f)
                evals[x] = {}
                for cv in cv_ids:
                    fn_eval = indir + x + "_" + cv + "_evaluation.pickle"
                    with open(fn_eval, "rb") as f:
                        evals[x][cv] = pickle.load(f)
                histories[x] = {}
                for cv in cv_ids:
                    fn_eval = indir + x + "_" + cv + "_history.pickle"
                    with open(fn_eval, "rb") as f:
                        histories[x][cv] = pickle.load(f)
            self.runparams[gs_id] = runparams
            self.run_ids_clean[gs_id] = run_ids_clean
            self.evals[gs_id] = evals
            self.histories[gs_id] = histories
            self.cv_ids[gs_id] = cv_ids
            for run_id in run_ids_clean:
                self.source_gs[run_id] = gs_id
            if len(run_ids_clean) == 0:
                raise ValueError("no complete runs found")

            # Summarise all metrics in a single table with rows for each runs and CV partition.
            params_list = list(
                {
                    "data_set": [runparams[x]["data_set"] for x in run_ids_clean],
                    "model_class": [runparams[x]["model_class"] for x in run_ids_clean],
                    "cond_type": [runparams[x]["cond_type"] for x in run_ids_clean]
                    if "cond_type" in list(runparams[x].keys())
                    else "none",
                    "gs_id": [runparams[x]["gs_id"] for x in run_ids_clean],
                    "model_id": [runparams[x]["model_id"] for x in run_ids_clean],
                    "split_mode": [runparams[x]["split_mode"] for x in run_ids_clean],
                    "radius": [runparams[x]["max_dist"] for x in run_ids_clean],
                    "graph_covar_selection": [runparams[x]["graph_covar_selection"] for x in run_ids_clean],
                    "node_label_space_id": [runparams[x]["node_feature_space_id_0"] for x in run_ids_clean],
                    "node_feature_space_id": [runparams[x]["node_feature_space_id_1"] for x in run_ids_clean],
                    "feature_transformation": [runparams[x]["feature_transformation"] for x in run_ids_clean],
                    "use_covar_node_position": [runparams[x]["use_covar_node_position"] for x in run_ids_clean],
                    "use_covar_node_label": [runparams[x]["use_covar_node_label"] for x in run_ids_clean],
                    "use_covar_graph_covar": [runparams[x]["use_covar_graph_covar"] for x in run_ids_clean],
                    "target_cell_type": [runparams[x]["target_cell_type"] for x in run_ids_clean]
                    if "target_cell_type" in list(runparams[x].keys())
                    else "none",
                    "optimizer": [runparams[x]["optimizer"] for x in run_ids_clean],
                    "learning_rate": [runparams[x]["learning_rate"] for x in run_ids_clean],
                    "intermediate_dim_enc": [runparams[x]["intermediate_dim_enc"] for x in run_ids_clean]
                    if "intermediate_dim_enc" in list(runparams[x].keys())
                    else "none",
                    "intermediate_dim_dec": [runparams[x]["intermediate_dim_dec"] for x in run_ids_clean]
                    if "intermediate_dim_dec" in list(runparams[x].keys())
                    else "none",
                    "latent_dim": [runparams[x]["latent_dim"] for x in run_ids_clean]
                    if "latent_dim" in list(runparams[x].keys())
                    else "none",
                    "depth_enc": [runparams[x]["depth_enc"] for x in run_ids_clean]
                    if "depth_enc" in list(runparams[x].keys())
                    else "none",
                    "depth_dec": [runparams[x]["depth_dec"] for x in run_ids_clean]
                    if "depth_dec" in list(runparams[x].keys())
                    else "none",
                    "dropout_rate": [runparams[x]["dropout_rate"] for x in run_ids_clean]
                    if "dropout_rate" in list(runparams[x].keys())
                    else "none",
                    "l2_coef": [runparams[x]["l2_coef"] for x in run_ids_clean]
                    if "l2_coef" in list(runparams[x].keys())
                    else "none",
                    "l1_coef": [runparams[x]["l1_coef"] for x in run_ids_clean]
                    if "l1_coef" in list(runparams[x].keys())
                    else "none",
                    "pretrain_decoder": [runparams[x]["pretrain_decoder"] for x in run_ids_clean]
                    if "pretrain_decoder" in list(runparams[x].keys())
                    else "none",
                    "aggressive": [runparams[x]["aggressive"] for x in run_ids_clean]
                    if "aggressive" in list(runparams[x].keys())
                    else "none",
                    "cond_depth": [str(runparams[x]["cond_depth"]) for x in run_ids_clean]
                    if "cond_depth" in list(runparams[x].keys())
                    else "none",
                    "cond_dim": [str(runparams[x]["cond_dim"]) for x in run_ids_clean]
                    if "cond_dim" in list(runparams[x].keys())
                    else "none",
                    "cond_dropout_rate": [str(runparams[x]["cond_dropout_rate"]) for x in run_ids_clean]
                    if "cond_dropout_rate" in list(runparams[x].keys())
                    else "none",
                    "cond_activation": [str(runparams[x]["cond_activation"]) for x in run_ids_clean]
                    if "cond_activation" in list(runparams[x].keys())
                    else "none",
                    "cond_l2_reg": [str(runparams[x]["cond_l2_reg"]) for x in run_ids_clean]
                    if "cond_l2_reg" in list(runparams[x].keys())
                    else "none",
                    "cond_use_bias": [str(runparams[x]["cond_use_bias"]) for x in run_ids_clean]
                    if "cond_bias" in list(runparams[x].keys())
                    else "none",
                    "use_domain": [runparams[x]["use_domain"] for x in run_ids_clean],
                    "domain_type": [runparams[x]["domain_type"] for x in run_ids_clean],
                    "use_batch_norm": [runparams[x]["use_batch_norm"] for x in run_ids_clean],
                    "use_type_cond": [runparams[x]["use_type_cond"] for x in run_ids_clean]
                    if "use_type_cond" in list(runparams[x].keys())
                    else "none",
                    "scale_node_size": [runparams[x]["scale_node_size"] for x in run_ids_clean],
                    "transform_input": [runparams[x]["transform_input"] for x in run_ids_clean],
                    "output_layer": [runparams[x]["output_layer"] for x in run_ids_clean],
                    "log_transform": [runparams[x]["log_transform"] for x in run_ids_clean]
                    if "log_transform" in list(runparams[x].keys())
                    else False,
                    "epochs": [runparams[x]["epochs"] for x in run_ids_clean],
                    "batch_size": [runparams[x]["batch_size"] for x in run_ids_clean],
                    "n_eval_nodes_per_graph": [runparams[x]["n_eval_nodes_per_graph"] for x in run_ids_clean],
                    "run_id_params": run_ids_clean,
                }.items()
            )
            self.runparams_table.append(pd.DataFrame(dict(params_list)))

            self.summary_table.append(
                pd.concat(
                    [
                        pd.DataFrame(
                            dict(
                                list(params_list)
                                + list(
                                    {
                                        "cv": cv,
                                        "run_id": run_ids_clean,
                                        "model": "_".join(gs_id.split("/")[-1].split("_")[1:-1]),
                                        "forward_pass_eval": True,
                                    }.items()
                                )
                                + list(
                                    dict(
                                        [
                                            (
                                                "train_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                [evals[x][cv]["train"][m] for x in run_ids_clean],
                                            )
                                            for m in list(evals[run_ids_clean[0]][cv]["train"].keys())
                                        ]
                                    ).items()
                                )
                                + list(
                                    dict(
                                        [
                                            (
                                                "val_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                [evals[x][cv]["val"][m] for x in run_ids_clean],
                                            )
                                            for m in list(evals[run_ids_clean[0]][cv]["val"].keys())
                                        ]
                                    ).items()
                                )
                                + list(
                                    dict(
                                        [
                                            (
                                                "test_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                [evals[x][cv]["test"][m] for x in run_ids_clean],
                                            )
                                            for m in list(evals[run_ids_clean[0]][cv]["test"].keys())
                                        ]
                                    ).items()
                                )
                                + list(
                                    dict(
                                        [
                                            (
                                                "all_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                [evals[x][cv]["all"][m] for x in run_ids_clean],
                                            )
                                            for m in list(evals[run_ids_clean[0]][cv]["all"].keys())
                                        ]
                                    ).items()
                                )
                            )
                        )
                        for cv in cv_ids
                    ]
                )
            )
            print("%s: loaded %i runs with %i-fold cross validation" % (gs_id, len(run_ids_clean), len(cv_ids)))

            if runparams[run_ids_clean[0]]["model_class"] in ["vae", "cvae", "cvae_ncem"]:
                evals_posterior_sampling = {}  # Dictionary over runs with dictionary over cross-validations with
                # results from model neighbourhood transfer evaluation.
                for x in run_ids_clean:
                    evals_posterior_sampling[x] = {}
                    for cv in cv_ids:
                        if add_posterior_sampling_model:
                            fn_eval = indir + x + "_" + cv + "_evaluation_posterior_sampling.pickle"
                            with open(fn_eval, "rb") as f:
                                evals_posterior_sampling[x][cv] = pickle.load(f)
                self.evals_posterior_sampling[gs_id + "_POSTERIOR_SAMPLING"] = evals_posterior_sampling

                if add_posterior_sampling_model:
                    self.summary_table.append(
                        pd.concat(
                            [
                                pd.DataFrame(
                                    dict(
                                        list(params_list)
                                        + list(
                                            {
                                                "cv": cv,
                                                "run_id": [x + "_posterior_sampling" for x in run_ids_clean],
                                                "model": "_".join(gs_id.split("/")[-1].split("_")[1:-1])
                                                + "_POSTERIOR_SAMPLING",
                                                "posterior_sampling_eval": True,
                                            }.items()
                                        )
                                        + list(
                                            dict(
                                                [
                                                    (
                                                        "train_"
                                                        + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                        [
                                                            evals_posterior_sampling[x][cv]["train"][m]
                                                            for x in run_ids_clean
                                                        ],
                                                    )
                                                    for m in list(
                                                        evals_posterior_sampling[run_ids_clean[0]][cv]["train"].keys()
                                                    )
                                                ]
                                            ).items()
                                        )
                                        + list(
                                            dict(
                                                [
                                                    (
                                                        "val_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                        [
                                                            evals_posterior_sampling[x][cv]["val"][m]
                                                            for x in run_ids_clean
                                                        ],
                                                    )
                                                    for m in list(
                                                        evals_posterior_sampling[run_ids_clean[0]][cv]["val"].keys()
                                                    )
                                                ]
                                            ).items()
                                        )
                                        + list(
                                            dict(
                                                [
                                                    (
                                                        "test_"
                                                        + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                        [
                                                            evals_posterior_sampling[x][cv]["test"][m]
                                                            for x in run_ids_clean
                                                        ],
                                                    )
                                                    for m in list(
                                                        evals_posterior_sampling[run_ids_clean[0]][cv]["test"].keys()
                                                    )
                                                ]
                                            ).items()
                                        )
                                        + list(
                                            dict(
                                                [
                                                    (
                                                        "all_" + m.replace("reconstruction_", "").replace("logp1_", ""),
                                                        [
                                                            evals_posterior_sampling[x][cv]["all"][m]
                                                            for x in run_ids_clean
                                                        ],
                                                    )
                                                    for m in list(
                                                        evals_posterior_sampling[run_ids_clean[0]][cv]["all"].keys()
                                                    )
                                                ]
                                            ).items()
                                        )
                                    )
                                )
                                for cv in cv_ids
                            ]
                        )
                    )
                    print(
                        "%s: loaded posterior sampling as seperate model (%i runs with %i-fold cross validation)"
                        % (gs_id, len(run_ids_clean), len(cv_ids))
                    )

        self.summary_table = pd.concat(self.summary_table)
        self.runparams_table = pd.concat(self.runparams_table)
        self.summary_table["um_radius"] = (self.summary_table["radius"] * self.lateral_resolution).astype(int)

        # Load files that are shared across a grid search.
        keys_all = {}
        keys_test = {}
        keys_train = {}
        keys_val = {}
        nodes_test = {}
        nodes_train = {}
        nodes_val = {}
        for cv in cv_ids:
            with open(indir + run_ids_clean[0] + "_" + cv + "_indices.pickle", "rb") as f:
                idx_partitions = pickle.load(f)
            keys_all[cv] = idx_partitions["all"]
            keys_test[cv] = idx_partitions["test"]
            keys_train[cv] = idx_partitions["train"]
            keys_val[cv] = idx_partitions["val"]
            nodes_test[cv] = idx_partitions["test_nodes"]
            nodes_train[cv] = idx_partitions["train_nodes"]
            nodes_val[cv] = idx_partitions["val_nodes"]
        self.keys_all = keys_all
        self.keys_test = keys_test
        self.keys_train = keys_train
        self.keys_val = keys_val
        self.nodes_test = nodes_test
        self.nodes_train = nodes_train
        self.nodes_val = nodes_val

    def load_target_cell_evaluation(self, report_unsuccessful_runs: bool = False):
        self.target_cell_runparams = {}
        self.target_cell_evals = {}
        self.target_cell_indices = {}
        target_cell_table = []
        for gs_id in self.gs_ids:
            # Collect runs that belong to grid search by looping over file names in directory.
            indir = self.source_path + gs_id + "/results/"
            # runs_ids are the unique hyper-parameter settings, which are again subsetted by cross-validation.
            # These ids are present in all files names but are only collected from the *model.tf file names here.

            run_ids = np.sort(
                np.unique(
                    [
                        "_".join(".".join(x.split(".")[:-1]).split("_")[:-1])
                        for x in os.listdir(indir)
                        if x.split("_")[-1].split(".")[0] == "time"
                    ]
                )
            )
            cv_ids = np.sort(np.unique([x.split("_")[-1] for x in run_ids]))  # identifiers of cross-validation splits
            run_ids = np.sort(
                np.unique(["_".join(x.split("_")[:-1]) for x in run_ids])  # identifiers of hyper-parameters settings
            )
            run_ids_clean = []  # only IDs of completed runs (all files present)
            for j, r in enumerate(run_ids):
                complete_run = True
                for cv in cv_ids:
                    fn = r + "_" + cv + "_ntevaluation.pickle"
                    if not os.path.isfile(indir + fn):
                        if report_unsuccessful_runs:
                            print("File %r missing" % fn)
                        complete_run = False
                # Check run parameter files (one per cross-validation set):
                fn = r + "_runparams.pickle"
                if not os.path.isfile(indir + fn):
                    if report_unsuccessful_runs:
                        print("File %r missing" % fn)
                    complete_run = False
                if not complete_run:
                    if report_unsuccessful_runs:
                        print("Run %r not successful" % r + "_" + cv)
                else:
                    run_ids_clean.append(r)
            # Load results and settings from completed runs:
            evals = (
                {}
            )  # Dictionary over runs with dictionary over cross-validations with results from model evaluation.
            indices = {}
            runparams = {}  # Dictionary over runs with model settings.
            for x in run_ids_clean:
                # Load model settings (these are shared across all partitions).
                fn_runparams = indir + x + "_runparams.pickle"
                with open(fn_runparams, "rb") as f:
                    runparams[x] = pickle.load(f)
                evals[x] = {}
                for cv in cv_ids:
                    fn_eval = indir + x + "_" + cv + "_ntevaluation.pickle"
                    with open(fn_eval, "rb") as f:
                        evals[x][cv] = pickle.load(f)
                indices[x] = {}
                for cv in cv_ids:
                    fn_eval = indir + x + "_" + cv + "_ntindices.pickle"
                    with open(fn_eval, "rb") as f:
                        indices[x][cv] = pickle.load(f)
            self.target_cell_runparams[gs_id] = runparams
            self.target_cell_evals[gs_id] = evals
            self.target_cell_indices[gs_id] = indices

            for x in run_ids_clean:
                for cv in cv_ids:
                    for target_cell in evals[x][cv].keys():
                        if evals[x][cv][target_cell]:
                            tc_frequencies = {
                                k: len(indices[x][cv][target_cell]["nodes_idx"][k])
                                for k in indices[x][cv][target_cell]["nodes_idx"].keys()
                            }

                            target_cell_table.append(
                                pd.concat(
                                    [
                                        pd.DataFrame(
                                            dict(
                                                list(
                                                    {
                                                        "model_class": [runparams[x]["model_class"]],
                                                        "cond_type": [runparams[x]["cond_type"]]
                                                        if "cond_type" in list(runparams[x].keys())
                                                        else "none",
                                                        "gs_id": [runparams[x]["gs_id"]],
                                                        "model_id": [runparams[x]["model_id"]],
                                                        "split_mode": [runparams[x]["split_mode"]],
                                                        "radius": [runparams[x]["max_dist"]],
                                                        "graph_covar_selection": [
                                                            runparams[x]["graph_covar_selection"]
                                                        ],
                                                        "node_label_space_id": [
                                                            runparams[x]["node_feature_space_id_0"]
                                                        ],
                                                        "node_feature_space_id": [
                                                            runparams[x]["node_feature_space_id_1"]
                                                        ],
                                                        "feature_transformation": [
                                                            runparams[x]["feature_transformation"]
                                                        ],
                                                        "use_covar_node_position": [
                                                            runparams[x]["use_covar_node_position"]
                                                        ],
                                                        "use_covar_node_label": [runparams[x]["use_covar_node_label"]],
                                                        "use_covar_graph_covar": [
                                                            runparams[x]["use_covar_graph_covar"]
                                                        ],
                                                        # "hold_out_covariate": [runparams[x]['hold_out_covariate']],
                                                        "optimizer": [runparams[x]["optimizer"]],
                                                        "learning_rate": [runparams[x]["learning_rate"]],
                                                        "intermediate_dim_enc": [runparams[x]["intermediate_dim_enc"]],
                                                        "intermediate_dim_dec": [runparams[x]["intermediate_dim_dec"]],
                                                        "latent_dim": [runparams[x]["latent_dim"]],
                                                        # "depth_enc": [runparams[x]['depth_enc']],
                                                        # "depth_dec": [runparams[x]['depth_dec']],
                                                        "dropout_rate": [runparams[x]["dropout_rate"]],
                                                        "l2_coef": [runparams[x]["l2_coef"]],
                                                        "l1_coef": [runparams[x]["l1_coef"]],
                                                        "use_domain": [runparams[x]["use_domain"]],
                                                        "domain_type": [runparams[x]["domain_type"]],
                                                        "use_batch_norm": [runparams[x]["use_batch_norm"]],
                                                        "scale_node_size": [runparams[x]["scale_node_size"]],
                                                        "transform_input": [runparams[x]["transform_input"]],
                                                        "output_layer": [runparams[x]["output_layer"]],
                                                        "log_transform": [runparams[x]["log_transform"]]
                                                        if "log_transform" in list(runparams[x].keys())
                                                        else False,
                                                        "epochs": [runparams[x]["epochs"]],
                                                        "batch_size": [runparams[x]["batch_size"]],
                                                        "run_id": x,
                                                        "cv": cv,
                                                        "target_cell_frequencies": sum(tc_frequencies.values()),
                                                        "model": "_".join(gs_id.split("/")[-1].split("_")[1:-1]),
                                                        "target_cell": target_cell,
                                                        "loss": [evals[x][cv][target_cell]["loss"]]
                                                        if "loss" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "custom_mean_sd": [evals[x][cv][target_cell]["custom_mean_sd"]]
                                                        if "custom_mean_sd" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "custom_mse": [evals[x][cv][target_cell]["custom_mse"]]
                                                        if "custom_mse" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "custom_mse_scaled": [
                                                            evals[x][cv][target_cell]["custom_mse_scaled"]
                                                        ]
                                                        if "custom_mse_scaled" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "gaussian_reconstruction_loss": [
                                                            evals[x][cv][target_cell]["gaussian_reconstruction_loss"]
                                                        ]
                                                        if "gaussian_reconstruction_loss"
                                                        in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "r_squared": [evals[x][cv][target_cell]["r_squared"]]
                                                        if "r_squared" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                        "r_squared_linreg": [
                                                            evals[x][cv][target_cell]["r_squared_linreg"]
                                                        ]
                                                        if "r_squared_linreg" in list(evals[x][cv][target_cell].keys())
                                                        else np.nan,
                                                    }.items()
                                                )
                                            )
                                        )
                                    ]
                                )
                            )
        self.target_cell_table = pd.concat(target_cell_table)
        self.target_cell_table["um_radius"] = (self.target_cell_table["radius"] * self.lateral_resolution).astype(int)

    def select_cv(self, cv_idx: int) -> str:
        """
        Return key of of cross-validation selected with numeric index.
        :param cv_idx: Index of cross-validation to plot confusion matrix for.
        :return:
        """
        if cv_idx >= len(self.cv_keys):
            raise ValueError("cv_idx %i out of scope of cross-validation set: %s" % (cv_idx, self.cv_keys))
        else:
            cv = self.cv_keys[cv_idx]
        print("cross-validation selected: %s" % cv)
        return cv

    def get_best_model_id(
        self,
        subset_hyperparameters: Optional[List[Tuple[str, str]]] = None,
        metric_select: str = "r_squared_linreg",
        partition_select: str = "test",
        cv_mode: str = "mean",
    ):
        if subset_hyperparameters is None:
            subset_hyperparameters = []
        if metric_select.endswith("loss"):
            ascending = True
            if cv_mode == "max":
                raise Warning("selected cv_mode max with metric_id loss, likely not intended")
        elif metric_select.endswith("elbo"):
            ascending = True
            if cv_mode == "max":
                raise Warning("selected cv_mode max with metric_id elbo, likely not intended")
        elif metric_select.endswith("mse") or metric_select.endswith("mse_scaled"):
            ascending = True
            if cv_mode == "max":
                raise Warning("selected cv_mode max with metric_id mse, likely not intended")
        elif metric_select.endswith("mae"):
            ascending = True
            if cv_mode == "max":
                raise Warning("selected cv_mode max with metric_id mae, likely not intended")
        elif metric_select.endswith("r_squared") or metric_select.endswith("r_squared_linreg"):
            ascending = False
        else:
            raise ValueError("measure %s not recognized" % metric_select)
        if partition_select not in ["test", "val", "train", "all"]:
            raise ValueError("partition %s not recognised" % partition_select)
        metric_select = partition_select + "_" + metric_select
        summary_table = self.summary_table
        for x, y in subset_hyperparameters:
            if not isinstance(y, list):
                y = [y]
            if not np.any([xx in y for xx in summary_table[x].values]):
                print(
                    "subset was empty, available values for %s are %s, given was %s"
                    % (x, str(np.unique(summary_table[x].values).tolist()), str(y))
                )
            summary_table = summary_table.loc[[xx in y for xx in summary_table[x].values], :]
        if cv_mode.lower() == "mean":
            best_model = (
                summary_table.groupby("run_id", as_index=False)[metric_select]
                .mean()
                .sort_values([metric_select], ascending=ascending)
            )
        elif cv_mode.lower() == "median":
            best_model = (
                summary_table.groupby("run_id", as_index=False)[metric_select]
                .median()
                .sort_values([metric_select], ascending=ascending)
            )
        elif cv_mode.lower() == "max":
            best_model = (
                summary_table.groupby("run_id", as_index=False)[metric_select]
                .max()
                .sort_values([metric_select], ascending=ascending)
            )
        elif cv_mode.lower() == "min":
            best_model = (
                summary_table.groupby("run_id", as_index=False)[metric_select]
                .min()
                .sort_values([metric_select], ascending=ascending)
            )
        else:
            raise ValueError("cv_mode %s not recognized" % cv_mode)
        return best_model["run_id"].values[0] if best_model.shape[0] > 0 else None

    def copy_best_model(
        self,
        gs_id: str,
        dst: str = "best",
        metric_select: str = "loss",
        partition_select: str = "val",
        cv_mode: str = "mean",
    ):
        from shutil import copyfile

        if dst[0] != "/":
            dst = self.source_path + gs_id + "/" + dst + "/"
        run_id = self.get_best_model_id(metric_select=metric_select, partition_select=partition_select, cv_mode=cv_mode)
        cvs = self.summary_table.loc[self.summary_table["run_id"].values == run_id, "cv"].values
        src = self.source_path + gs_id + "/results/"
        print("copying model files from %s to %s" % (src, dst))
        for cv in cvs:
            fn_model = run_id + "_" + cv + "_model.h5"
            copyfile(src + fn_model, dst + fn_model)
            fn_idx = run_id + "_" + cv + "_indices.pickle"
            copyfile(src + fn_idx, dst + fn_idx)

    def get_info(self, model_id, expected_pickle: Union[None, list] = None):
        if expected_pickle is None:
            expected_pickle = ["evaluation", "history", "hyperparam"]

        indir = self.source_path + self.source_gs[model_id] + "/results/"
        # Check that all files are present:
        cv_ids = np.sort(
            np.unique([x.split("_")[-2] for x in os.listdir(indir) if x.split("_")[-1].split(".")[0] == "time"])
        )
        for cv in cv_ids:
            # Check pickled files:
            for suffix in expected_pickle:
                fn = model_id + "_" + cv + "_" + suffix + ".pickle"
                if not os.path.isfile(indir + fn):
                    raise ValueError("file %s missing" % suffix)
        info = {}
        for cv in cv_ids:
            info[cv] = {}
            for suffix in expected_pickle:
                fn = model_id + "_" + cv + "_" + suffix + ".pickle"
                with open(indir + fn, "rb") as f:
                    info[cv][suffix] = pickle.load(f)
        fn = model_id + "_runparams.pickle"
        with open(indir + fn, "rb") as f:
            info["runparams"] = pickle.load(f)
        self.info = info

        # Plotting functions:

    def plot_best_model_by_hyperparam(
        self,
        partition_show: str,
        metric_show: str,
        partition_select: str,
        metric_select: str,
        param_x: str,
        param_hue: str,
        graph_model_class: str,
        baseline_model_class: str,
        rename_levels: Optional[List[Tuple[str, str]]] = None,
        subset_hyperparameters: Optional[List[Tuple[str, str]]] = None,
        cv_mode: str = "mean",
        yaxis_limit: Optional[Tuple[float, float]] = None,
        xticks: Optional[List[int]] = None,
        rotate_xticks: bool = True,
        figsize: Tuple[float, float] = (3.5, 4.0),
        fontsize: Optional[int] = None,
        plot_mode: str = "boxplot",
        hue_order=None,
        palette: Optional[dict] = None,
        color: Optional[str] = None,
        save: Optional[str] = None,
        suffix: str = "best_by_hyperparam.pdf",
        show: bool = True,
        return_axs: bool = False,
    ):
        if fontsize:
            sc.set_figure_params(scanpy=True, fontsize=fontsize)
        sns.set_palette("colorblind")
        rename_levels = [] if not rename_levels else rename_levels
        subset_hyperparameters = [] if not subset_hyperparameters else subset_hyperparameters

        plt.ioff()
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        if isinstance(param_hue, list):
            param_hue_new = param_hue[0] + "_" + param_hue[1]
            self.summary_table[param_hue_new] = [
                str(self.summary_table[param_hue[0]].values[i]) + "_" + str(self.summary_table[param_hue[1]].values[i])
                for i in range(self.summary_table.shape[0])
            ]
            param_hue = param_hue_new
        params_x_unique = np.sort(np.unique(self.summary_table[param_x].values))
        params_hue_unique = np.sort(np.unique(self.summary_table[param_hue].values))

        run_ids = []
        summary_table = self.summary_table.copy()
        runparams_table = self.runparams_table.copy()
        for x in params_x_unique:
            for hue in params_hue_unique:
                run_id_temp = self.get_best_model_id(
                    subset_hyperparameters=[(param_x, x), (param_hue, hue)] + subset_hyperparameters,
                    partition_select=partition_select,
                    metric_select=metric_select,
                    cv_mode=cv_mode,
                )
                # print(run_id_temp)
                if run_id_temp is not None:
                    run_ids.append(run_id_temp)

        summary_table = summary_table.loc[np.array([x in run_ids for x in summary_table["run_id"].values]), :]
        runparams_table = runparams_table.loc[
            np.array([x in run_ids for x in runparams_table["run_id_params"].values]), :
        ]
        for level, d in rename_levels:
            summary_table[level] = [d[x] for x in summary_table[level].values]
        self.best_model_hyperparam_table = runparams_table
        summary_table.sort_values([param_x, param_hue])
        ycol = partition_show + "_" + metric_show

        if plot_mode == "boxplot":
            if color:
                sns.boxplot(x=param_x, y=ycol, data=summary_table, ax=ax, color=color)
                sns.swarmplot(
                    x=param_x,
                    y=ycol,
                    palette=[color],
                    data=summary_table,
                    ax=ax,
                )
            else:
                sns.boxplot(
                    x=param_x,
                    y=ycol,
                    hue=param_hue,
                    hue_order=hue_order,
                    data=summary_table,
                    ax=ax,
                )
                sns.swarmplot(x=param_x, y=ycol, hue=param_hue, data=summary_table, ax=ax)
        elif plot_mode == "lineplot":
            sns.scatterplot(
                x=param_x,
                y=ycol,
                hue=param_hue,
                style="cv",
                palette=palette,
                data=summary_table[summary_table["model_class"] == baseline_model_class],
                s=100,
                ax=ax,
            )
            sns.lineplot(
                x=param_x,
                y=ycol,
                style="cv",
                hue_order=hue_order,
                palette=palette,
                data=summary_table[summary_table["model_class"] == graph_model_class],
                hue=param_hue,
                ax=ax,
                markers=True,
            )
            ax.set_xscale("symlog")
        elif plot_mode == "mean_lineplot":
            temp_summary_table = summary_table[[param_hue, param_x, ycol]].groupby([param_hue, param_x]).mean()
            sns.lineplot(
                x=param_x,
                y=ycol,
                palette=["grey"],
                style=param_hue,
                data=temp_summary_table,
                hue=param_hue,
                ax=ax,
                sort=False,
                markers=True,
            )
            ax.set_xscale("symlog")

        ax.grid(False)
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        # Put a legend to the right of the current axis
        lgd = ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_title("")
        if yaxis_limit is not None:
            ax.set_ylim(yaxis_limit)
        if xticks is not None:
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticks)
        ax.yaxis.set_major_formatter(FormatStrFormatter("%0.3f"))

        if rotate_xticks:
            plt.xticks(rotation=90)

        # Save, show and return figure.
        # plt.tight_layout()
        if save is not None:
            plt.savefig(save + "_" + partition_show + suffix, bbox_extra_artists=(lgd,), bbox_inches="tight")

        if show:
            plt.show()

        plt.close(fig)
        plt.ion()

        if return_axs:
            return ax
        else:
            return None

    def plot_target_cell_evaluation(
        self,
        metric_show: str,
        metric_select: str,
        param_x: str,
        ncols: int = 8,
        plot_mode: str = "boxplot",
        show: bool = True,
        save: Optional[str] = None,
        suffix: str = "target_cell_evaluation.pdf",
        return_axs: bool = False,
        panel_width: float = 3.0,
        panel_height: float = 3.0,
    ):
        params_x_unique = np.sort(np.unique(self.target_cell_table[param_x].values))
        params_tc_unique = np.sort(np.unique(self.target_cell_table["target_cell"].values))

        ct = len(params_tc_unique)
        nrows = len(params_tc_unique) // ncols + int(len(params_tc_unique) % ncols > 0)

        fig, ax = plt.subplots(
            ncols=ncols, nrows=nrows, figsize=(ncols * panel_width, nrows * panel_height), sharex=True
        )
        ax = ax.flat
        for a in ax[ct:]:
            a.remove()
        ax = ax[:ct]
        ax = ax.ravel()
        target_cell_frequencies = {}
        for i, tc in enumerate(params_tc_unique):
            run_ids = []
            for x in params_x_unique:

                target_cell_table = self.target_cell_table[self.target_cell_table["target_cell"] == tc]
                subset_hyperparameters = [(param_x, x)]
                for a, b in subset_hyperparameters:
                    if not isinstance(b, list):
                        b = [b]
                    if not np.any([xx in b for xx in target_cell_table[a].values]):
                        print(
                            "subset was empty, available values for %s are %s, given was %s"
                            % (a, str(np.unique(target_cell_table[a].values).tolist()), str(b))
                        )
                    summary_table = target_cell_table.loc[[xx in b for xx in target_cell_table[a].values], :]

                best_model = (
                    summary_table.groupby("run_id", as_index=False)[metric_select]
                    .mean()
                    .sort_values([metric_select], ascending=False, na_position="last")
                )
                run_id_temp = best_model["run_id"].values[0]

                if run_id_temp is not None:
                    run_ids.append(run_id_temp)

            table = target_cell_table.loc[np.array([x in run_ids for x in target_cell_table["run_id"].values]), :]
            tc_frequencies = int(np.unique(table["target_cell_frequencies"])[0])
            target_cell_frequencies.update({tc: tc_frequencies})
            if plot_mode == "boxplot":
                sns.boxplot(x=param_x, y=metric_show, data=table, ax=ax[i], color="steelblue")
                sns.swarmplot(x=param_x, y=metric_show, data=table, ax=ax[i], color="steelblue")
            elif plot_mode == "lineplot":
                sns.lineplot(x=param_x, y=metric_show, style="cv", data=table, ax=ax[i], sort=False, markers=True)
            ax[i].set_title(f"{tc.replace('_', ' ')} \n({str(tc_frequencies)} cells)")
            ax[i].set_xlabel("")
            ax[i].set_ylabel("")
            ax[i].yaxis.set_major_formatter(FormatStrFormatter("%0.3f"))
            for tick in ax[i].get_xticklabels():
                tick.set_rotation(90)

        self.target_cell_frequencies = target_cell_frequencies
        plt.tight_layout()
        # Save, show and return figure.
        if save is not None:
            plt.savefig(save + "_" + suffix)

        if show:
            plt.show()

        plt.close(fig)
        plt.ion()

        if return_axs:
            return ax
        else:
            return None



import pickle
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import tensorflow as tf
from anndata import AnnData
from diffxpy.testing.correction import correct
from scipy import sparse, stats
from scipy.sparse import csr_matrix
from tqdm import tqdm

import ncem.estimators as estimators
import ncem.models as models
import ncem.train as train
from ncem.utils.wald_test import wald_test


class InterpreterBase(estimators.Estimator):
    data_path: str
    results_path: str

    model_class: str
    data_set: str
    radius: int

    cell_names: list

    def __init__(self):
        super().__init__()
        self.cell_type = None
        self.position_matrix = None
        self.adj_type = "full"

        self._model_kwargs = None
        self._fn_model_weights = None
        self.gscontainer = None
        self.model_id = None
        self.gs_id = None
        self.gscontainer_runparams = None

        self.reinit_model = None

    def init_model(self):
        raise ValueError("models should not be initialized within interpreter class, use load_model()")

    def load_model(
        self,
        results_path: str,
        gs_id: str,
        cv_idx: int,
        subset_hyperparameters: Optional[List[Tuple[str, str]]] = None,
        model_id: Optional[str] = None,
        expected_pickle: Optional[list] = None,
        lateral_resolution: float = 1.0,
    ):
        """
        Load best or selected model from grid search directory.

        :param results_path:
        :param gs_id:
        :param cv_idx:
        :param subset_hyperparameters:
        :param model_id:
        :param expected_pickle:
        :param lateral_resolution:
        :return:
        """
        if subset_hyperparameters is None:
            subset_hyperparameters = []
        if expected_pickle is None:
            expected_pickle = ["evaluation", "history", "hyperparam", "model_args", "time"]

        gscontainer = train.GridSearchContainer(results_path, gs_id, lateral_resolution)
        gscontainer.load_gs(expected_pickle=expected_pickle)
        if model_id is None:
            model_id = gscontainer.get_best_model_id(
                subset_hyperparameters=subset_hyperparameters,
                metric_select="loss",
                cv_mode="mean",
                partition_select="test",
            )
        cv = gscontainer.select_cv(cv_idx=cv_idx)
        fn_model_kwargs = f"{results_path}{gs_id}/results/{model_id}_{cv}_model_args.pickle"
        fn_weights = f"{results_path}{gs_id}/results/{model_id}_{cv}_model_weights.tf"
        with open(fn_model_kwargs, "rb") as f:
            model_kwargs = pickle.load(f)

        self._model_kwargs = model_kwargs
        self._fn_model_weights = fn_weights
        self.gscontainer = gscontainer
        self.model_id = model_id
        self.gs_id = gs_id
        self.gscontainer_runparams = self.gscontainer.runparams[self.gs_id][self.model_id]
        self.results_path = results_path
        print("loaded model %s" % model_id)

    def _get_dataset(
        self,
        image_keys: List[str],
        nodes_idx: Dict[str, np.ndarray],
        batch_size: int,
        shuffle_buffer_size: int,
        train: bool,
        seed: Optional[int],
        prefetch: int = 100,
        reinit_n_eval: Optional[int] = None,
    ):
        pass

    def _get_resampled_dataset(
        self, image_keys: np.ndarray, nodes_idx: dict, batch_size: int, seed: Optional[int] = None, prefetch: int = 100
    ):
        pass

    def get_data_again(self, data_path: str, data_origin: str):
        """
        Loads data as previously done during model training.

        :param data_path:
        :param data_origin:
        :return:
        """
        self.cond_type = self.gscontainer_runparams["cond_type"] if "cond_type" in self.gscontainer_runparams else None
        if self.cond_type == "gcn":
            self.adj_type = "scaled"
        elif self.cond_type in ["max", None]:
            self.adj_type = "full"
        else:
            raise ValueError("cond_type %s not recognized" % self.cond_type)

        self.get_data(
            data_origin=data_origin,
            data_path=data_path,
            radius=int(self.gscontainer_runparams["max_dist"]),
            graph_covar_selection=self.gscontainer_runparams["graph_covar_selection"],
            node_label_space_id=self.gscontainer_runparams["node_feature_space_id_0"],
            node_feature_space_id=self.gscontainer_runparams["node_feature_space_id_1"],
            use_covar_node_position=self.gscontainer_runparams["use_covar_node_position"],
            use_covar_node_label=self.gscontainer_runparams["use_covar_node_label"],
            use_covar_graph_covar=self.gscontainer_runparams["use_covar_graph_covar"],
            domain_type=self.gscontainer_runparams["domain_type"],
        )
        self.data_path = data_path
        self.n_eval_nodes_per_graph = self.gscontainer_runparams["n_eval_nodes_per_graph"]
        self.model_class = self.gscontainer_runparams["model_class"]
        self.data_set = self.gscontainer_runparams["data_set"]
        self.radius = int(self.gscontainer_runparams["max_dist"])
        self.cond_depth = (
            self.gscontainer_runparams["cond_depth"] if "cond_depth" in self.gscontainer_runparams.keys() else None
        )
        self.log_transform = self.gscontainer_runparams["log_transform"]
        self.cell_names = list(self.node_type_names.values())

    def split_data_byidx_again(self, cv_idx: int):
        """
        Split data into partitions as done during model training.

        :param cv_idx: Index of cross-validation to plot confusion matrix for.
        :return:
        """
        cv = self.gscontainer.select_cv(cv_idx=cv_idx)
        fn = f"{self.results_path}{self.gs_id}/results/{self.model_id}_{cv}_indices.pickle"
        with open(fn, "rb") as f:
            indices = pickle.load(f)

        self.split_data_given(
            img_keys_test=indices["test"],
            img_keys_train=indices["train"],
            img_keys_eval=indices["val"],
            nodes_idx_test=indices["test_nodes"],
            nodes_idx_train=indices["train_nodes"],
            nodes_idx_eval=indices["val_nodes"],
        )

    def init_model_again(self):
        if self.model_class in ["vae"]:
            model = models.ModelCVAE(**self._model_kwargs)
        elif self.model_class in ["cvae", "cvae_ncem"]:
            model = models.ModelCVAEncem(**self._model_kwargs)
        elif self.model_class == ["ed", "lvmnp"]:
            model = models.ModelED(**self._model_kwargs)
        elif self.model_class in ["ed_ncem", "clvmnp"]:
            model = models.ModelEDncem(**self._model_kwargs)
        elif self.model_class in ["linear", "linear_baseline"]:
            model = models.ModelLinear(**self._model_kwargs)
        elif self.model_class in ["interactions", "interactions_baseline"]:
            model = models.ModelInteractions(**self._model_kwargs)
        else:
            raise ValueError("model_class not recognized")
        self.model = model

        self.vi_model = False  # variational inference
        if self.model_class in ["vae", "cvae", "cvae_ncem"]:
            self.vi_model = True

    def load_weights_again(self):
        self.model.training_model.load_weights(self._fn_model_weights)

    def reinitialize_model(self, changed_model_kwargs: dict, print_summary: bool = False):
        assert self.model is not None, "no model loaded, run init_model_again() first"
        # updating new model kwargs
        new_model_kwargs = self._model_kwargs.copy()
        new_model_kwargs.update(changed_model_kwargs)

        if self.model_class in ["vae"]:
            reinit_model = models.ModelCVAE(**new_model_kwargs)
        elif self.model_class in ["cvae", "cvae_ncem"]:
            reinit_model = models.ModelCVAEncem(**new_model_kwargs)
        elif self.model_class == ["ed", "lvmnp"]:
            reinit_model = models.ModelED(**new_model_kwargs)
        elif self.model_class in ["ed_ncem", "clvmnp"]:
            reinit_model = models.ModelEDncem(**new_model_kwargs)
        elif self.model_class in ["linear", "linear_baseline"]:
            reinit_model = models.ModelLinear(**new_model_kwargs)
        elif self.model_class in ["interactions", "interactions_baseline"]:
            reinit_model = models.ModelInteractions(**new_model_kwargs)
        else:
            raise ValueError("model_class not recognized")
        self.reinit_model = reinit_model

        if print_summary:
            self.reinit_model.training_model.summary()
        print(f"setting reinitialized layer weights to layer weights from model {self.model_id}")
        for layer, new_layer in zip(self.model.training_model.layers, self.reinit_model.training_model.layers):
            new_layer.set_weights(layer.get_weights())

    def _pp_saliencies(
        self, gradients, h_0, h_0_full, remove_own_gradient: bool = True, absolute_saliencies: bool = True
    ):
        if self.cond_type == "max":
            gradients = np.nan_to_num(gradients)
        # removing gradient self node type
        if remove_own_gradient:
            identity = np.ones(shape=gradients.shape)
            for i in range(gradients.shape[0]):
                identity[i, :] = h_0
            gradients = gradients * (np.ones(shape=gradients.shape) - identity)
        if absolute_saliencies:
            gradients = np.abs(gradients)
        gradients = np.sum(gradients, axis=-1, dtype=np.float64)
        gradients = np.expand_dims(gradients, axis=0)

        sal = np.matmul(gradients, h_0_full)  # 1 x max_nodes
        return sal

    def _neighbourhood_frequencies(self, a, h_0_full, discretize_adjacency: bool = True):
        neighbourhood = []
        for i, adj in enumerate(a):
            if discretize_adjacency:
                adj = np.asarray(adj > 0, dtype="int")
            neighbourhood.append(np.matmul(adj, h_0_full[i].astype(int)))  # 1 x node_types
        neighbourhood = pd.DataFrame(np.concatenate(neighbourhood, axis=0), columns=self.cell_names, dtype=np.float)
        return neighbourhood


class InterpreterLinear(estimators.EstimatorLinear, InterpreterBase):
    """
    Inherits all relevant functions specific to EstimatorLinear estimators
    """

    def __init__(self):
        super().__init__()

    def _get_np_data(
        self, image_keys: Union[np.ndarray, str], nodes_idx: Union[dict, str]
    ) -> Tuple[Tuple[list, list, list, list, list], list]:
        """
        :param image_keys: Observation images indices.
        :param nodes_idx: Observation nodes indices.
        :return: Tuple of list of raw data, one list entry per image / graph
        """
        if isinstance(image_keys, (int, np.int32, np.int64)):
            image_keys = [image_keys]
        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        # Loop over sub-selected data set and sum gradients across all selected observations.
        target = []
        source = []
        sf = []
        node_covar = []
        g = []
        h_obs = []

        for step, (x_batch, y_batch) in enumerate(ds):
            target_batch, source_batch, sf_batch, node_covar_batch, g_batch = x_batch
            target.append(target_batch.numpy().squeeze())
            source.append(source_batch.numpy().squeeze())
            sf.append(sf_batch.numpy().squeeze())
            node_covar.append(node_covar_batch.numpy().squeeze())
            g.append(g_batch.numpy().squeeze())
            h_obs.append(y_batch[0].numpy().squeeze())
        return (target, source, sf, node_covar, g), h_obs


class InterpreterInteraction(estimators.EstimatorInteractions, InterpreterBase):
    """
    Inherits all relevant functions specific to EstimatorInteractions estimators
    """

    def __init__(self):
        super().__init__()

    def _get_np_data(
        self, image_keys: Union[np.ndarray, str], nodes_idx: Union[dict, str]
    ) -> Tuple[Tuple[list, List[csr_matrix], list, list, list], list]:
        """
        :param image_keys: Observation images indices.
        :param nodes_idx: Observation nodes indices.
        :return: Tuple of list of raw data, one list entry per image / graph
        """
        if isinstance(image_keys, (int, np.int32, np.int64)):
            image_keys = [image_keys]
        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        # Loop over sub-selected data set and sum gradients across all selected observations.
        target = []
        interactions = []
        sf = []
        node_covar = []
        g = []
        h_obs = []

        for step, (x_batch, y_batch) in enumerate(ds):
            target_batch, interaction_batch, sf_batch, node_covar_batch, g_batch = x_batch
            target.append(target_batch.numpy().squeeze())
            interactions.append(
                csr_matrix(
                    (
                        interaction_batch.values.numpy(),
                        (interaction_batch.indices.numpy()[:, 1], interaction_batch.indices.numpy()[:, 2]),
                    ),
                    shape=interaction_batch.dense_shape.numpy()[1:],
                ).todense()
            )
            sf.append(sf_batch.numpy().squeeze())
            node_covar.append(node_covar_batch.numpy().squeeze())
            g.append(g_batch.numpy().squeeze())
            h_obs.append(y_batch[0].numpy().squeeze())
        return (target, interactions, sf, node_covar, g), h_obs

    def target_cell_relative_performance(
        self,
        image_key: str,
        baseline_model,
        target_cell_type: str,
        undefined_type: Optional[str] = None,
        n_neighbors: Optional[int] = None,
        n_pcs: Optional[int] = None,
    ):
        cluster_col = self.data.celldata.uns["metadata"]["cluster_col_preprocessed"]
        if isinstance(image_key, str):
            image_key = [image_key]
        adata_list = []
        tqdm_total = 0
        for key in image_key:
            tqdm_total = tqdm_total + len(self.nodes_idx_all[str(key)])
        with tqdm(total=tqdm_total) as pbar:
            for key in image_key:
                nodes_idx = {str(key): self.nodes_idx_all[str(key)]}
                ds = self._get_dataset(
                    image_keys=[str(key)],
                    nodes_idx=nodes_idx,
                    batch_size=1,
                    shuffle_buffer_size=1,
                    train=False,
                    seed=None,
                    reinit_n_eval=1,
                )
                graph = []
                baseline = []
                for step, (x_batch, y_batch) in enumerate(ds):
                    out_graph = self.reinit_model.training_model(x_batch)
                    out_graph = np.split(ary=out_graph.numpy().squeeze(), indices_or_sections=2, axis=-1)[0]

                    out_base = baseline_model.reinit_model.training_model(x_batch)
                    out_base = np.split(ary=out_base.numpy().squeeze(), indices_or_sections=2, axis=-1)[0]

                    r2_graph = stats.linregress(out_graph, y_batch.numpy().squeeze())[2] ** 2
                    graph.append(r2_graph)

                    r2_base = stats.linregress(out_base, y_batch.numpy().squeeze())[2] ** 2
                    baseline.append(r2_base)
                    pbar.update(1)
                temp_adata = self.data.img_celldata[key].copy()
                if undefined_type:
                    temp_adata = temp_adata[temp_adata.obs[cluster_col] != undefined_type]
                temp_adata.obs["relative_r_squared"] = np.array(graph) - np.array(baseline)
                adata_list.append(temp_adata)

        adata = adata_list[0].concatenate(adata_list[1:], uns_merge="same") if len(adata_list) > 0 else adata_list
        adata_tc = adata[(adata.obs[cluster_col] == target_cell_type)].copy()
        sc.pp.neighbors(adata_tc, n_neighbors=n_neighbors, n_pcs=n_pcs)
        sc.tl.louvain(adata_tc)
        sc.tl.umap(adata_tc)

        print("n cells: ", adata_tc.shape[0])
        adata_tc.obs[f"{target_cell_type} substates"] = f"{target_cell_type} " + adata_tc.obs.louvain.astype(str)
        adata_tc.obs[f"{target_cell_type} substates"] = adata_tc.obs[f"{target_cell_type} substates"].astype("category")

        print(adata_tc.obs[f"{target_cell_type} substates"].value_counts())

        return adata, adata_tc

    @staticmethod
    def plot_substate_performance(
        adata,
        target_cell_type: str,
        relative_performance_key: str = "relative_r_squared",
        fontsize: Optional[int] = None,
        figsize: Tuple[float, float] = (4.0, 4.0),
        palette: Optional[list] = None,
        save: Optional[str] = None,
        suffix: str = "_substates_performance.pdf",
        show: bool = True,
        return_axs: bool = False,
    ):
        if palette is None:
            palette = ["#1f77b4", "#2ca02c", "#8c564b", "#7f7f7f", "#17becf"]
        if fontsize:
            sc.set_figure_params(scanpy=True, fontsize=fontsize)
        plt.rcParams["axes.grid"] = False
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        sns.boxplot(
            data=adata.obs,
            x=f"{target_cell_type} substates",
            y=relative_performance_key,
            order=list(np.unique(adata.obs[f"{target_cell_type} substates"])),
            palette=palette,
            ax=ax,
        )
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.tick_params(axis="x", labelrotation=90)

        # Save, show and return figure.
        plt.tight_layout()
        if save is not None:
            plt.savefig(save + target_cell_type + suffix)

        if show:
            plt.show()

        plt.close(fig)
        plt.ion()

        if return_axs:
            return ax
        else:
            return None

    def plot_spatial_relative_performance(
        self,
        adata,
        target_cell_type: str,
        relative_performance_key: str = "relative_r_squared",
        figsize: Tuple[float, float] = (7.0, 5.0),
        fontsize: Optional[int] = None,
        spot_size: int = 40,
        clean_view: bool = False,
        save: Optional[str] = None,
        suffix: str = "_spatial_relative_performance.pdf",
        show: bool = True,
        return_axs: bool = False,
    ):
        class MidpointNormalize(colors.Normalize):
            def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
                self.midpoint = midpoint
                colors.Normalize.__init__(self, vmin, vmax, clip)

            def __call__(self, value, clip=None):
                x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
                return np.ma.masked_array(np.interp(value, x, y))

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        if fontsize:
            sc.set_figure_params(scanpy=True, fontsize=fontsize)

        if clean_view:
            adata = adata[np.argwhere(np.array(adata.obsm["spatial"])[:, 1] < 0).squeeze()]
        sc.pl.spatial(
            adata[adata.obs[adata.uns["metadata"]["cluster_col_preprocessed"]] != target_cell_type],
            spot_size=spot_size,
            ax=ax,
            show=False,
            na_color="whitesmoke",
            title="",
        )
        sc.pl.spatial(
            adata[adata.obs[adata.uns["metadata"]["cluster_col_preprocessed"]] == target_cell_type],
            color=relative_performance_key,
            spot_size=spot_size,
            cmap="coolwarm",
            norm=MidpointNormalize(midpoint=0.0),
            ax=ax,
            show=False,
            title="",
        )

        ax.invert_yaxis()
        # Save, show and return figure.
        plt.tight_layout()
        if save is not None:
            plt.savefig(save + target_cell_type + suffix)

        if show:
            plt.show()

        plt.close(fig)
        plt.ion()

        if return_axs:
            return ax
        else:
            return None

    def relative_performance_grid(
        self,
        image_keys: Union[np.ndarray, str],
        nodes_idx: Union[dict, str],
        base_interpreter,
        scale_node_frequencies: int,
        metric: str = "r_squared_linreg",
        mode: str = "mean",
        figsize: Tuple[float, float] = (6.0, 5.0),
        save: Optional[str] = None,
        suffix: str = "_expression_grid.pdf",
        show: bool = True,
    ):
        class MidpointNormalize(colors.Normalize):
            def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
                self.midpoint = midpoint
                colors.Normalize.__init__(self, vmin, vmax, clip)

            def __call__(self, value, clip=None):
                x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
                return np.ma.masked_array(np.interp(value, x, y))

        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        node_names = list(self.node_type_names.values())
        interaction_names = []
        for source in node_names:
            for target in node_names:
                interaction_names.append(f"{target} - {source}")
        expression = []
        target = []
        interactions = []

        base_pred = []
        graph_pred = []
        for step, (x_batch, y_batch) in enumerate(ds):
            target_batch, interactions_batch, sf_batch, node_covar_batch, g_batch = x_batch

            base_pred.append(np.squeeze(base_interpreter.model.training_model(x_batch)[0].numpy()))
            graph_pred.append(np.squeeze(self.model.training_model(x_batch)[0].numpy()))
            expression.append(y_batch[0].numpy().squeeze())
            target.append(target_batch.numpy().squeeze())
            interactions.append(
                sparse.csr_matrix(
                    (
                        interactions_batch.values.numpy(),
                        (interactions_batch.indices.numpy()[:, 1], interactions_batch.indices.numpy()[:, 2]),
                    ),
                    shape=interactions_batch.dense_shape.numpy()[1:],
                ).todense()
            )

        interactions = pd.DataFrame(np.concatenate(interactions, axis=0), columns=interaction_names)
        expression = np.concatenate(expression, axis=0)
        base_pred = np.split(ary=np.concatenate(base_pred, axis=0), indices_or_sections=2, axis=1)[0]
        graph_pred = np.split(ary=np.concatenate(graph_pred, axis=0), indices_or_sections=2, axis=1)[0]

        plt.ioff()
        # function returns a n_features_type x n_features_type grid
        grid_summary = []
        for i, k in enumerate(interaction_names):
            temp_interactions = interactions[interactions[k] > 0][k]

            if mode == "mean":
                true = np.mean(expression[list(temp_interactions.index), :], axis=0)
                base = np.mean(base_pred[list(temp_interactions.index), :], axis=0)
                graph = np.mean(graph_pred[list(temp_interactions.index), :], axis=0)
            elif mode == "var":
                true = np.var(expression[list(temp_interactions.index), :], axis=0)
                base = np.var(base_pred[list(temp_interactions.index), :], axis=0)
                graph = np.var(graph_pred[list(temp_interactions.index), :], axis=0)
            else:
                raise ValueError(f"{mode} not implemented")

            if metric == "r_squared_linreg":
                base_metric = stats.linregress(true, base)[2] ** 2
                graph_metric = stats.linregress(true, graph)[2] ** 2
            elif metric == "mae":
                base_metric = np.mean(np.abs(base - true))
                graph_metric = np.mean(np.abs(graph - true))
            else:
                raise ValueError(f"{metric} not implemented")

            target = k.split(" - ")[0]
            source = k.split(" - ")[1]

            grid_summary.append(
                np.array(
                    [target, source, np.sum(np.array(temp_interactions), dtype=np.int32), base_metric, graph_metric]
                )
            )

        expression_grid_summary = np.concatenate(np.expand_dims(grid_summary, axis=0), axis=0)
        plt.ioff()
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        temp_df = (
            pd.DataFrame(
                expression_grid_summary, columns=["target", "source", "contact_frequency", "baseline", "graph"]
            )
            .astype({"baseline": "float32", "graph": "float32", "contact_frequency": "int32"})
            .sort_values(["target", "source"], ascending=[False, True])
        )
        if metric == "r_squared_linreg":
            temp_df["relative_performance"] = temp_df.graph - temp_df.baseline
            metric_name = "R2 (linear regression)"
        elif metric == "mae":
            temp_df["relative_performance"] = temp_df.baseline - temp_df.graph
            metric_name = "MAE"
        else:
            raise ValueError(f"{metric} not implemented")
        img0 = ax.scatter(
            x=temp_df.source,
            y=temp_df.target,
            s=temp_df.contact_frequency / scale_node_frequencies,
            c=temp_df.relative_performance,
            cmap="seismic",
            norm=MidpointNormalize(midpoint=0.0),
        )
        cbar = plt.colorbar(img0, ax=ax)
        cbar.set_label(f"relative {metric_name}", rotation=90)
        ax.tick_params(axis="x", labelrotation=90)
        plt.tight_layout()
        if save is not None:
            plt.savefig(save + "_" + metric + suffix)
        if show:
            plt.show()
        plt.close(fig)
        plt.ion()

    def interaction_significance(
        self,
        image_keys,
        nodes_idx,
        significance_threshold: float = 0.01,
    ):
        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        interaction_shape = self.model.training_model.inputs[1].shape
        interaction_params = (
            self.model.training_model.weights[0]
            .numpy()[self.n_features_0 : interaction_shape[-1] + self.n_features_0, :]
            .T
        )

        interaction_params = np.concatenate(
            np.expand_dims(np.split(interaction_params, indices_or_sections=self.n_features_0, axis=1), axis=-1),
            axis=-1,
        )

        interactions = []
        y = []
        for step, (x_batch, y_batch) in enumerate(ds):
            target_batch, interactions_batch, sf_batch, node_covar_batch, g_batch = x_batch
            interactions.append(
                sparse.csr_matrix(
                    (
                        interactions_batch.values.numpy(),
                        (interactions_batch.indices.numpy()[:, 1], interactions_batch.indices.numpy()[:, 2]),
                    ),
                    shape=interactions_batch.dense_shape.numpy()[1:],
                ).todense()
            )
            y.append(y_batch[0].numpy().squeeze())

        interactions = np.concatenate(interactions, axis=0)
        y = np.concatenate(y, axis=0)

        interactions = np.concatenate(
            np.expand_dims(np.split(interactions, indices_or_sections=self.n_features_0, axis=1), axis=-1), axis=-1
        )

        fisher_inv = []
        for i in range(interactions.shape[1]):
            target_x = interactions[:, i, :]
            target_fisher_inv = np.divide(
                np.expand_dims(np.matmul(target_x.T, target_x), axis=0),
                np.expand_dims(np.expand_dims(np.var(y, axis=0), axis=-1), axis=-1),
            )
            fisher_inv.append(target_fisher_inv)

        bool_significance, significance = wald_test(
            parameters=interaction_params, fisher_inv=fisher_inv, significance_threshold=significance_threshold
        )
        return interaction_params, bool_significance, significance


class InterpreterGraph(estimators.EstimatorGraph, InterpreterBase):
    """
    Inherits all relevant functions specific to EstimatorGraph estimators
    """

    def __init__(self):
        super().__init__()

    def _get_np_data(
        self, image_keys: Union[np.ndarray, str], nodes_idx: Union[dict, str]
    ) -> Tuple[Tuple[list, list, list, list, List[csr_matrix], List[csr_matrix], list, list], list]:
        """
        :param image_keys: Observation images indices.
        :param nodes_idx: Observation nodes indices.
        :return: Tuple of list of raw data, one list entry per image / graph
        """
        if isinstance(image_keys, (int, np.int32, np.int64)):
            image_keys = [image_keys]
        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        # Loop over sub-selected data set and sum gradients across all selected observations.
        h_1 = []
        sf = []
        h_0 = []
        h_0_full = []
        a = []
        a_full = []
        node_covar = []
        g = []
        h_obs = []

        for step, (x_batch, y_batch) in enumerate(ds):
            h_1_batch, sf_batch, h_0_batch, h_0_full_batch, a_batch, a_full_batch, node_covar_batch, g_batch = x_batch
            h_1.append(h_1_batch.numpy().squeeze())
            sf.append(sf_batch.numpy().squeeze())
            h_0.append(h_0_batch.numpy().squeeze())
            h_0_full.append(h_0_full_batch.numpy().squeeze())
            a.append(
                csr_matrix(
                    (a_batch.values.numpy(), (a_batch.indices.numpy()[:, 1], a_batch.indices.numpy()[:, 2])),
                    shape=a_batch.dense_shape.numpy()[1:],
                )
            )
            a_full.append(
                csr_matrix(
                    (
                        a_full_batch.values.numpy().squeeze(),
                        (a_full_batch.indices.numpy().squeeze()[:, 1], a_full_batch.indices.numpy().squeeze()[:, 2]),
                    ),
                    shape=a_full_batch.dense_shape.numpy().squeeze()[1:],
                )
            )
            node_covar.append(node_covar_batch.numpy().squeeze())
            g.append(g_batch.numpy().squeeze())
            h_obs.append(y_batch[0].numpy().squeeze())
        return (h_1, sf, h_0, h_0_full, a, a_full, node_covar, g), h_obs


class InterpreterEDncem(estimators.EstimatorEDncem, InterpreterGraph):
    """
    Inherits all relevant functions specific to EstimatorInteractions estimators
    """

    def __init__(self):
        super().__init__()

    def target_cell_saliencies(
        self,
        target_cell_type: str,
        drop_columns: Optional[list] = None,
        dop_images: Optional[list] = None,
        partition: str = "test",
        multicolumns: Optional[list] = None,
    ):
        target_cell_idx = list(self.node_type_names.values()).index(target_cell_type)

        if partition == "test":
            img_keys = self.img_keys_test
            node_idxs = self.nodes_idx_test
        elif partition == "train":
            img_keys = self.img_keys_train
            node_idxs = self.nodes_idx_train
        elif partition == "val":
            img_keys = self.img_keys_eval
            node_idxs = self.nodes_idx_eval
        elif partition == "all":
            img_keys = self.img_keys_all
            node_idxs = self.nodes_idx_all
        else:
            print("unknown partition selected, saliencies are computed on test partition")
            img_keys = self.img_keys_test
            node_idxs = self.nodes_idx_test

        img_saliency = []
        keys = []
        with tqdm(total=len(img_keys)) as pbar:
            for key in img_keys:
                idx = {key: node_idxs[key]}
                ds = self._get_dataset(
                    image_keys=[key],
                    nodes_idx=idx,
                    batch_size=1,
                    shuffle_buffer_size=1,
                    train=False,
                    seed=None,
                    reinit_n_eval=1,
                )
                # inputs extracted for plotting aggregation
                saliencies = []
                h_1 = []
                h_0 = []
                h_0_full = []
                a = []
                for step, (x_batch, y_batch) in enumerate(ds):
                    h_1_batch = x_batch[0].numpy().squeeze()
                    h_0_batch = x_batch[2].numpy().squeeze()  # 1 x 1 x node_types
                    h_0_full_batch = x_batch[3]  # 1 x max_nodes x node_types
                    a_batch = x_batch[4]  # 1 x 1 x max_nodes

                    if h_0_batch[target_cell_idx] == 1.0:
                        h_1.append(h_1_batch)
                        h_0.append(h_0_batch)
                        h_0_full.append(h_0_full_batch.numpy().squeeze())
                        a.append(
                            sparse.csr_matrix(
                                (
                                    a_batch.values.numpy(),
                                    (a_batch.indices.numpy()[:, 1], a_batch.indices.numpy()[:, 2]),
                                ),
                                shape=a_batch.dense_shape.numpy()[1:],
                            ).toarray()
                        )

                        # gradients with target h_0_full_batch
                        with tf.GradientTape(persistent=True) as tape:
                            tape.watch([h_0_full_batch])
                            model_out = self.reinit_model.training_model(x_batch)[0]  # 1 x max_nodes x node_features
                        grads = tape.gradient(model_out, h_0_full_batch)[0].numpy()  # 1 x max_nodes x node_types
                        grads = self._pp_saliencies(
                            gradients=grads,
                            h_0=h_0_batch,
                            h_0_full=h_0_full_batch.numpy().squeeze(),
                            remove_own_gradient=True,
                            absolute_saliencies=False,
                        )
                        saliencies.append(grads)
                if len(saliencies) == 0:
                    continue
                saliencies = np.concatenate(saliencies, axis=0)
                saliencies = np.sum(saliencies, axis=0)
                neighbourhood = self._neighbourhood_frequencies(a=a, h_0_full=h_0_full, discretize_adjacency=True)
                neighbourhood = np.sum(np.array(neighbourhood), axis=0)
                img_saliency.append(saliencies / neighbourhood)
                keys.append(key)
                pbar.update(1)
        columns = (
            [(x.replace("_", " ").split()[0], x.replace("_", " ").split()[1]) for x in keys] if multicolumns else keys
        )

        df = pd.DataFrame(
            np.concatenate(np.expand_dims(img_saliency, axis=0), axis=1).T,
            columns=columns,
            index=list(self.node_type_names.values()),
        )
        df.columns = pd.MultiIndex.from_tuples(df.columns, names=multicolumns) if multicolumns else df.columns

        df = df.reindex(sorted(df.columns), axis=1)

        if drop_columns:
            df = df.drop(drop_columns)
        if dop_images:
            df = df.drop(dop_images, axis=1, level=1)

        return df

    @staticmethod
    def plot_target_cell_saliencies(
        saliencies,
        multiindex: bool = True,
        fontsize: Optional[int] = None,
        figsize: Tuple[float, float] = (30.0, 7.0),
        width_ratios: Optional[list] = None,
        save: Optional[str] = None,
        suffix: str = "_imagewise_target_cell_saliencies.pdf",
        show: bool = True,
        return_axs: bool = False,
    ):
        if width_ratios is None:
            width_ratios = [5, 1]
        if fontsize:
            sc.set_figure_params(scanpy=True, fontsize=fontsize)
        plt.rcParams["axes.grid"] = False
        plt.ioff()
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=figsize, gridspec_kw={"width_ratios": width_ratios})
        sns.heatmap(saliencies, cmap="seismic", ax=ax[0], center=0)
        ax[0].set_xlabel("")

        if multiindex:
            xlabel_mapping = OrderedDict()
            for index1, index2 in saliencies.columns:
                xlabel_mapping.setdefault(index1, [])
                xlabel_mapping[index1].append(index2.replace("slice", ""))

            hline = []
            new_xlabels = []
            for index1, index2_list in xlabel_mapping.items():
                index2_list[0] = "{}".format(index2_list[0])
                new_xlabels.extend(index2_list)

                if hline:
                    hline.append(len(index2_list) + hline[-1])
                else:
                    hline.append(len(index2_list))
            ax[0].set_xticklabels(new_xlabels, rotation=90)

        sns.boxplot(data=saliencies.T, ax=ax[1], orient="h", color="steelblue", showfliers=True)
        ax[1].set_yticklabels("")

        # Save, show and return figure.
        plt.tight_layout()
        if save is not None:
            plt.savefig(save + suffix)
        if show:
            plt.show()
        plt.close(fig)
        plt.ion()
        return ax if return_axs else None


class InterpreterCVAEncem(estimators.EstimatorCVAEncem, InterpreterGraph):
    """
    Inherits all relevant functions specific to EstimatorInteractions estimators
    """

    def __init__(self):
        super().__init__()

    def compute_latent_space_cluster_enrichment(
        self,
        image_key,
        target_cell_type: str,
        n_neighbors: Optional[int] = None,
        n_pcs: Optional[int] = None,
        filter_titles: Optional[List[str]] = None,
        clip_pvalues: Optional[int] = -5,
    ):

        node_type_names = list(self.data.celldata.uns["node_type_names"].values())
        ds = self._get_dataset(
            image_keys=[image_key],
            nodes_idx={image_key: self.nodes_idx_all[image_key]},
            batch_size=1,
            shuffle_buffer_size=1,
            seed=None,
            train=False,
            reinit_n_eval=1,
        )

        cond_encoder = tf.keras.Model(
            self.reinit_model.training_model.input, self.reinit_model.training_model.get_layer("cond_encoder").output
        )
        cond_latent_z_mean = []
        h_0_full = []
        h_0 = []
        a = []
        for step, (x_batch, y_batch) in enumerate(ds):
            h_1_batch, sf_batch, h_0_batch, h_0_full_batch, a_batch, a_full_batch, node_covar_batch, g_batch = x_batch

            cond_z, cond_z_mean, cond_z_log = cond_encoder(x_batch)
            cond_latent_z_mean.append(cond_z_mean.numpy())

            h_0.append(h_0_batch.numpy())
            h_0_full.append(h_0_full_batch.numpy().squeeze())
            a.append(
                sparse.csr_matrix(
                    (a_batch.values.numpy(), (a_batch.indices.numpy()[:, 1], a_batch.indices.numpy()[:, 2])),
                    shape=a_batch.dense_shape.numpy()[1:],
                ).toarray()
            )

        cond_latent_z_mean = np.concatenate(cond_latent_z_mean, axis=0)
        source_type = self._neighbourhood_frequencies(a=a, h_0_full=h_0_full, discretize_adjacency=True)
        source_type = (
            pd.DataFrame((source_type > 0).astype(str), columns=node_type_names)
            .replace({"True": "in neighbourhood", "False": "not in neighbourhood"}, regex=True)
            .astype("category")
        )
        h_0 = pd.DataFrame(np.concatenate(h_0, axis=0).squeeze(), columns=node_type_names)
        target_type = pd.DataFrame(np.array(h_0.idxmax(axis=1)), columns=["target_cell"])

        metadata = pd.concat([target_type, source_type], axis=1)
        cond_adata = AnnData(cond_latent_z_mean, obs=metadata)
        sc.pp.neighbors(cond_adata, n_neighbors=n_neighbors, n_pcs=n_pcs)
        sc.tl.louvain(cond_adata)
        sc.tl.umap(cond_adata)

        for i, x in enumerate(node_type_names):
            cond_adata.uns[f"{x}_colors"] = ["darkgreen", "lightgrey"]

        cond_adata.obs["substates"] = target_cell_type + " " + cond_adata.obs.louvain.astype(str)
        cond_adata.obs["substates"] = cond_adata.obs["substates"].astype("category")
        print("n cells: ", cond_adata.shape[0])
        substate_counts = cond_adata.obs["substates"].value_counts()
        print(substate_counts)

        one_hot = pd.get_dummies(cond_adata.obs.louvain, dtype=np.bool)
        # Join the encoded df
        df = cond_adata.obs.join(one_hot)

        distinct_louvain = len(np.unique(cond_adata.obs.louvain))
        pval_source_type = []
        for i, st in enumerate(node_type_names):
            pval_cluster = []
            for j in range(distinct_louvain):
                crosstab = np.array(pd.crosstab(df[f"{st}"], df[str(j)]))
                if crosstab.shape[0] < 2:
                    crosstab = np.vstack([crosstab, [0, 0]])
                oddsratio, pvalue = stats.fisher_exact(crosstab)
                pvalue = correct(np.array([pvalue]))
                pval_cluster.append(pvalue)
            pval_source_type.append(pval_cluster)

        columns = [f"{target_cell_type} {x}" for x in np.unique(cond_adata.obs.louvain)]
        pval = pd.DataFrame(np.array(pval_source_type).squeeze(), index=node_type_names, columns=columns)
        log_pval = np.log10(pval)

        if filter_titles:
            log_pval = log_pval.sort_values(columns, ascending=True).filter(items=filter_titles, axis=0)
        if clip_pvalues:
            log_pval[log_pval < clip_pvalues] = clip_pvalues

        fold_change_df = cond_adata.obs[["target_cell", "substates"] + node_type_names]
        counts = pd.pivot_table(
            fold_change_df.replace({"in neighbourhood": 1, "not in neighbourhood": 0}),
            index=["substates"],
            aggfunc=np.sum,
            margins=True,
        ).T

        fold_change = counts.loc[:, columns].div(np.array(substate_counts), axis=1)
        fold_change = fold_change.subtract(np.array(counts["All"] / cond_adata.shape[0]), axis=0)

        if filter_titles:
            fold_change = fold_change.fillna(0).filter(items=filter_titles, axis=0)
        return cond_adata.copy(), log_pval, fold_change


class InterpreterNoGraph(estimators.EstimatorNoGraph, InterpreterBase):
    """
    Inherits all relevant functions specific to EstimatorEDncem estimators and InterpreterBase
    """

    def __init__(self):
        super().__init__()

    def _get_np_data(
        self, image_keys: Union[np.ndarray, str], nodes_idx: Union[dict, str]
    ) -> Tuple[Tuple[list, list, list, list], list]:
        """
        :param image_keys: Observation images indices.
        :param nodes_idx: Observation nodes indices.
        :return: Tuple of list of raw data, one list entry per image / graph
        """
        if isinstance(image_keys, (int, np.int32, np.int64)):
            image_keys = [image_keys]
        ds = self._get_dataset(
            image_keys=image_keys, nodes_idx=nodes_idx, batch_size=1, shuffle_buffer_size=1, train=False, seed=None
        )
        # Loop over sub-selected data set and sum gradients across all selected observations.
        h_1 = []
        sf = []
        node_covar = []
        g = []
        h_obs = []

        for step, (x_batch, y_batch) in enumerate(ds):
            h_1_batch, sf_batch, node_covar_batch, g_batch = x_batch
            h_1.append(h_1_batch.numpy().squeeze())
            sf.append(sf_batch.numpy().squeeze())
            node_covar.append(node_covar_batch.numpy().squeeze())
            g.append(g_batch.numpy().squeeze())
            h_obs.append(y_batch[0].numpy().squeeze())
        return (h_1, sf, node_covar, g), h_obs



from typing import Union

import tensorflow as tf

from ncem.models.layers import (Decoder, Encoder, GaussianOutput, GCNLayer,
                                MaxLayer, NegBinConstDispOutput, NegBinOutput,
                                NegBinSharedDispOutput, NodeDegrees)


class ModelEDncem:
    def __init__(
        self,
        input_shapes,
        latent_dim: int = 10,
        dropout_rate: float = 0.1,
        l2_coef: float = 0.0,
        l1_coef: float = 0.0,
        enc_intermediate_dim: int = 128,
        enc_depth: int = 2,
        dec_intermediate_dim: int = 128,
        dec_depth: int = 2,
        cond_type: str = "gcn",
        cond_depth: int = 1,
        cond_dim: int = 8,
        cond_dropout_rate: float = 0.1,
        cond_activation: Union[str, tf.keras.layers.Layer] = "relu",
        cond_l2_reg: float = 0.0,
        cond_use_bias: bool = True,
        use_domain: bool = False,
        use_type_cond: bool = False,
        use_node_degree: bool = False,
        scale_node_size: bool = False,
        output_layer: str = "gaussian",
        log_transform: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.args = {
            "input_shapes": input_shapes,
            "latent_dim": latent_dim,
            "dropout_rate": dropout_rate,
            "l2_coef": l2_coef,
            "l1_coef": l1_coef,
            "enc_intermediate_dim": enc_intermediate_dim,
            "enc_depth": enc_depth,
            "dec_intermediate_dim": dec_intermediate_dim,
            "dec_depth": dec_depth,
            "cond_type": cond_type,
            "cond_depth": cond_depth,
            "cond_dim": cond_dim,
            "cond_dropout_rate": cond_dropout_rate,
            "cond_activation": cond_activation
            if not isinstance(cond_activation, tf.keras.layers.Layer)
            else cond_activation.name,
            "cond_l2_reg": cond_l2_reg,
            "cond_use_bias": cond_use_bias,
            "use_domain": use_domain,
            "use_type_cond": use_type_cond,
            "use_node_degree": use_node_degree,
            "scale_node_size": scale_node_size,
            "output_layer": output_layer,
            "log_transform": log_transform,
        }
        in_node_feature_dim = input_shapes[0]
        out_node_feature_dim = input_shapes[1]
        graph_dim = input_shapes[2]
        in_node_dim = input_shapes[3]
        categ_condition_dim = input_shapes[4]
        domain_dim = input_shapes[5]

        # node features - reconstruction: Input Tensor - shape=(None, N, F)
        input_x_reconstruct = tf.keras.Input(
            shape=(in_node_dim, out_node_feature_dim), name="node_features_reconstruct"
        )
        # node size - reconstruction: Input Tensor - shape=(None, N, 1)
        input_node_size = tf.keras.Input(shape=(in_node_dim, 1), name="node_size_reconstruct")
        # node features - node representation of other nodes: Input Tensor - shape=(None, N, F)
        input_x_cond = tf.keras.Input(shape=(in_node_dim, in_node_feature_dim), name="node_features_cond")
        # node features - node representation of other nodes: Input Tensor - shape=(None, N, F)
        input_x_cond_full = tf.keras.Input(shape=(graph_dim, in_node_feature_dim), name="node_features_cond_full")
        # adj_matrices - A: Input Tensor - shape=(None, N, N)
        input_a = tf.keras.Input(shape=(in_node_dim, graph_dim), name="adjacency_matrix", sparse=True)
        # full adj_matrices - A: Input Tensor - shape=(None, N, N)
        input_afull = tf.keras.Input(shape=(graph_dim, graph_dim), name="adjacency_matrix_full", sparse=True)
        # Categorical predictors: Input Tensor - shape=(None, N, P)
        input_categ_condition = tf.keras.Input(shape=(in_node_dim, categ_condition_dim), name="categorical_predictor")
        # domain information of graph - shape=(None, 1)
        input_g = tf.keras.layers.Input(shape=(domain_dim,), name="input_da_group", dtype="int32")

        # Decoder inputs:
        input_latent_sampling = tf.keras.Input(shape=(in_node_dim, latent_dim), name="z_sampling")
        latent_sampling_reshaped = tf.reshape(input_latent_sampling, [-1, latent_dim])

        if use_domain:
            categ_condition = tf.concat(
                [
                    input_categ_condition,
                    tf.tile(tf.expand_dims(tf.cast(input_g, dtype="float32"), axis=-2), [1, in_node_dim, 1]),
                ],
                axis=-1,
            )
        else:
            categ_condition = input_categ_condition

        if use_node_degree:
            node_degrees = NodeDegrees(in_node_dim)(input_a)

            categ_condition = tf.concat([categ_condition, node_degrees], axis=-1)

        if cond_depth > 1:
            print("using multi layer graph model")
        x_neighbour_embedding = input_x_cond_full
        if cond_type == "gcn" or cond_type == "gcn_full":
            for i in range(cond_depth - 1):
                cond_layer = GCNLayer(
                    output_dim=cond_dim,
                    dropout_rate=cond_dropout_rate,
                    activation=cond_activation,
                    l2_reg=cond_l2_reg,
                    use_bias=cond_use_bias,
                    padded=False,
                    name=f"conditional_layer_stack_{i}",
                )
                x_neighbour_embedding = cond_layer([x_neighbour_embedding, input_afull])
            # In last forward pass, only retain nodes that are to be decoded: not using the full adjacency matrix:
            cond_layer = GCNLayer(
                output_dim=cond_dim,
                dropout_rate=cond_dropout_rate,
                activation=cond_activation,
                l2_reg=cond_l2_reg,
                use_bias=cond_use_bias,
                padded=False,
                name=f"conditional_layer_stack_{cond_depth}",
            )
            x_neighbour_embedding = cond_layer([x_neighbour_embedding, input_a])
        elif cond_type == "max":
            for i in range(cond_depth - 1):
                cond_layer = MaxLayer(name=f"conditional_layer_stack_{i}")
                x_neighbour_embedding = cond_layer([x_neighbour_embedding, input_afull])
            # In last forward pass, only retain nodes that are to be decoded: not using the full adjacency matrix:
            cond_layer = MaxLayer(name=f"conditional_layer_stack_{cond_depth}")
            x_neighbour_embedding = cond_layer([x_neighbour_embedding, input_a])
        else:
            raise ValueError("tried to access a non-supported conditional layer %s" % cond_type)

        inputs_encoder = (x_neighbour_embedding, categ_condition)

        self.encoder_model = Encoder(
            latent_dim=latent_dim,
            intermediate_dim=enc_intermediate_dim,
            dropout_rate=dropout_rate,
            n_hidden=enc_depth,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            probabilistic=False,
            use_type_cond=use_type_cond,
        )
        output_encoder = self.encoder_model(inputs_encoder)

        z, z_mean, z_log_var = output_encoder

        self.decoder_model = Decoder(
            intermediate_dim=dec_intermediate_dim,
            dropout_rate=dropout_rate,
            n_hidden=dec_depth,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            use_type_cond=use_type_cond,
        )
        output_decoder = self.decoder_model((z, categ_condition))
        sampling_decoder = self.decoder_model((latent_sampling_reshaped, categ_condition))

        if output_layer == "gaussian":
            output_decoder_layer = GaussianOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="GaussianOutput_decoder",
            )((output_decoder, input_node_size))
            output_sampling_decoder = GaussianOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="GaussianOutput_sampling",
            )((sampling_decoder, input_node_size))
        elif output_layer == "nb":
            output_decoder_layer = NegBinOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinOutput_decoder",
            )((output_decoder, input_node_size))
            output_sampling_decoder = NegBinOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinOutput_sampling",
            )((sampling_decoder, input_node_size))
        elif output_layer == "nb_shared_disp":
            output_decoder_layer = NegBinSharedDispOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinSharedDispOutput_decoder",
            )((output_decoder, input_node_size))
            output_sampling_decoder = NegBinSharedDispOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinSharedDispOutput_sampling",
            )((sampling_decoder, input_node_size))
        elif output_layer == "nb_const_disp":
            output_decoder_layer = NegBinConstDispOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinConstDispOutput_decoder",
            )((output_decoder, input_node_size))
            output_sampling_decoder = NegBinConstDispOutput(
                original_dim=out_node_feature_dim,
                use_node_scale=scale_node_size,
                name="NegBinConstDispOutput_sampling",
            )((sampling_decoder, input_node_size))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)

        output_decoder_concat = tf.keras.layers.Concatenate(axis=2, name="reconstruction")(output_decoder_layer)
        output_sampling_concat = tf.keras.layers.Concatenate(axis=2, name="reconstruction")(output_sampling_decoder)

        self.encoder = tf.keras.Model(
            inputs=[
                input_x_reconstruct,
                input_x_cond,
                input_x_cond_full,
                input_a,
                input_afull,
                input_categ_condition,
                input_g,
            ],
            outputs=output_encoder,
            name="encoder_ncem",
        )
        self.decoder = tf.keras.Model(
            inputs=[
                input_latent_sampling,
                input_node_size,
                input_x_cond,
                input_x_cond_full,
                input_a,
                input_afull,
                input_categ_condition,
                input_g,
            ],
            outputs=output_sampling_concat,
            name="decoder_ncem",
        )
        self.training_model = tf.keras.Model(
            inputs=[
                input_x_reconstruct,
                input_node_size,
                input_x_cond,
                input_x_cond_full,
                input_a,
                input_afull,
                input_categ_condition,
                input_g,
            ],
            outputs=output_decoder_concat,
            name="ed_ncem",
        )



import tensorflow as tf

from ncem.estimators import EstimatorGraph
from ncem.models import ModelEDncem


class EstimatorEDncem(EstimatorGraph):
    def __init__(
        self,
        cond_type: str = "gcn",
        use_type_cond: bool = True,
        log_transform: bool = False,
    ):
        super(EstimatorEDncem, self).__init__()
        self.model_type = "ed_ncem"
        if cond_type == "gcn":
            self.adj_type = "scaled"
        elif cond_type == "max":
            self.adj_type = "full"
        else:
            raise ValueError("cond_type %s not recognized" % cond_type)
        self.cond_type = cond_type
        self.use_type_cond = use_type_cond
        self.log_transform = log_transform
        self.metrics = {"np": [], "tf": []}
        self.n_eval_nodes_per_graph = None
        self.cond_depth = None

    def init_model(
        self,
        optimizer: str = "adam",
        learning_rate: float = 0.0001,
        latent_dim: int = 10,
        dropout_rate: float = 0.1,
        l2_coef: float = 0.0,
        l1_coef: float = 0.0,
        enc_intermediate_dim: int = 128,
        enc_depth: int = 2,
        dec_intermediate_dim: int = 128,
        dec_depth: int = 2,
        cond_depth: int = 1,
        cond_dim: int = 8,
        cond_dropout_rate: float = 0.1,
        cond_activation: str = "relu",
        cond_l2_reg: float = 0.0,
        cond_use_bias: bool = False,
        n_eval_nodes_per_graph: int = 32,
        use_domain: bool = False,
        scale_node_size: bool = True,
        beta: float = 0.01,
        max_beta: float = 1.0,
        pre_warm_up: int = 0,
        output_layer: str = "gaussian",
        **kwargs
    ):
        self.n_eval_nodes_per_graph = n_eval_nodes_per_graph
        self.model = ModelEDncem(
            input_shapes=(
                self.n_features_0,
                self.n_features_1,
                self.max_nodes,
                self.n_eval_nodes_per_graph,
                self.n_node_covariates,
                self.n_domains,
            ),
            latent_dim=latent_dim,
            dropout_rate=dropout_rate,
            l2_coef=l2_coef,
            l1_coef=l1_coef,
            enc_intermediate_dim=enc_intermediate_dim,
            enc_depth=enc_depth,
            dec_intermediate_dim=dec_intermediate_dim,
            dec_depth=dec_depth,
            cond_type=self.cond_type,
            cond_depth=cond_depth,
            cond_dim=cond_dim,
            cond_dropout_rate=cond_dropout_rate,
            cond_activation=cond_activation,
            cond_l2_reg=cond_l2_reg,
            cond_use_bias=cond_use_bias,
            use_domain=use_domain,
            use_type_cond=self.use_type_cond,
            scale_node_size=scale_node_size,
            output_layer=output_layer,
        )
        optimizer = tf.keras.optimizers.get(optimizer)
        tf.keras.backend.set_value(optimizer.lr, learning_rate)
        self.cond_depth = cond_depth
        self.beta = beta
        self.max_beta = max_beta
        self.pre_warm_up = pre_warm_up
        self._compile_model(optimizer=optimizer, output_layer=output_layer)



from typing import Optional, Union

import numpy as np
import tensorflow as tf
from patsy import dmatrix
from scipy.sparse import coo_matrix

from ncem.estimators import Estimator
from ncem.models import ModelInteractions


class EstimatorInteractions(Estimator):
    def __init__(
        self,
        log_transform: bool = False,
    ):
        super(EstimatorInteractions, self).__init__()
        self.model_type = "linear_interactions"
        self.adj_type = "full"
        self.log_transform = log_transform
        self.metrics = {"np": [], "tf": []}
        self.n_eval_nodes_per_graph = None

    def _get_output_signature(self, resampled: bool = False):
        pass

    def _get_dataset(
        self,
        image_keys: np.ndarray,
        nodes_idx: Union[dict, str],
        batch_size: int,
        shuffle_buffer_size: int,
        train: bool,
        seed: Optional[int],
        prefetch: int = 100,
        reinit_n_eval: Optional[int] = None,
    ):
        np.random.seed(seed)
        if reinit_n_eval is not None and reinit_n_eval != self.n_eval_nodes_per_graph:
            print(
                "ATTENTION: specifying reinit_n_eval will change class argument n_eval_nodes_per_graph "
                "from %i to %i" % (self.n_eval_nodes_per_graph, reinit_n_eval)
            )
            self.n_eval_nodes_per_graph = reinit_n_eval

        def generator():
            for key in image_keys:
                if nodes_idx[key].size == 0:  # needed for images where no nodes are selected
                    continue
                idx_nodes = np.arange(0, self.a[key].shape[0])

                if train:
                    index_list = [
                        np.asarray(
                            np.random.choice(
                                a=nodes_idx[key],
                                size=self.n_eval_nodes_per_graph,
                                replace=True,
                            ),
                            dtype=np.int32,
                        )
                    ]
                else:
                    # dropping
                    index_list = [
                        np.asarray(
                            nodes_idx[key][self.n_eval_nodes_per_graph * i : self.n_eval_nodes_per_graph * (i + 1)],
                            dtype=np.int32,
                        )
                        for i in range(len(nodes_idx[key]) // self.n_eval_nodes_per_graph)
                    ]

                for indices in index_list:
                    h_1 = self.h_1[key][idx_nodes]
                    diff = self.max_nodes - h_1.shape[0]
                    zeros = np.zeros((diff, h_1.shape[1]))
                    h_1 = np.asarray(np.concatenate((h_1, zeros), axis=0), dtype="float32")
                    h_1 = h_1[indices]
                    if self.log_transform:
                        h_1 = np.log(h_1 + 1.0)

                    h_0 = self.h_0[key][idx_nodes]
                    diff = self.max_nodes - h_0.shape[0]
                    zeros = np.zeros((diff, h_0.shape[1]), dtype="float32")
                    h_0_full = np.asarray(np.concatenate((h_0, zeros), axis=0), dtype="float32")
                    h_0 = h_0_full[indices]

                    a = self.a[key][idx_nodes, :][:, idx_nodes]
                    a = a[indices, :]
                    coo = a.tocoo()
                    a_shape = np.asarray((self.n_eval_nodes_per_graph, self.max_nodes), dtype="int64")
                    a = coo_matrix((coo.data, (coo.row, coo.col)), a_shape)

                    # Assemble design matrix components for interactions:
                    dmat_neighbours = (
                        (a.toarray() > 0.0).astype("int").dot(h_0_full)
                    )  # n_eval_nodes_per_graph x node types
                    # discretize interactions
                    dmat_neighbours = np.asarray(dmat_neighbours > 0, dtype="int")

                    data = {"target": h_0, "dmat_neighbours": dmat_neighbours}
                    target = np.asarray(dmatrix("target-1", data))
                    interactions = np.asarray(dmatrix("target:dmat_neighbours-1", data))
                    # interactions_data = interactions.flatten()
                    # interactions_shape = np.asarray(
                    #    (self.n_eval_nodes_per_graph, self.n_features_0 ** 2), dtype="int64"
                    # )
                    interactions = coo_matrix(interactions)
                    coo = interactions.tocoo()
                    interactions_ind = np.asarray(np.mat([coo.row, coo.col]).transpose(), dtype="int64")
                    interactions_val = np.asarray(coo.data, dtype="float32")
                    interactions_shape = np.asarray(
                        (self.n_eval_nodes_per_graph, self.n_features_0 ** 2), dtype="int64"
                    )
                    interactions = tf.SparseTensor(
                        indices=interactions_ind, values=interactions_val, dense_shape=interactions_shape
                    )

                    node_covar = self.node_covar[key][idx_nodes]
                    diff = self.max_nodes - node_covar.shape[0]
                    zeros = np.zeros((diff, node_covar.shape[1]))
                    node_covar = np.asarray(np.concatenate([node_covar, zeros], axis=0), dtype="float32")
                    node_covar = node_covar[indices]

                    sf = np.expand_dims(self.size_factors[key][idx_nodes], axis=1)
                    diff = self.max_nodes - sf.shape[0]
                    zeros = np.zeros((diff, sf.shape[1]))
                    sf = np.asarray(np.concatenate([sf, zeros], axis=0), dtype="float32")
                    sf = sf[indices, :]

                    g = np.zeros((self.n_domains,), dtype="int32")
                    g[self.domains[key]] = 1

                    yield (target, interactions, sf, node_covar, g), h_1

        dataset = tf.data.Dataset.from_generator(
            generator=generator,
            output_signature=(
                (
                    tf.TensorSpec(shape=(self.n_eval_nodes_per_graph, self.n_features_0), dtype=tf.float32),
                    tf.SparseTensorSpec(shape=(self.n_eval_nodes_per_graph, self.n_features_0 ** 2), dtype=tf.float32),
                    tf.TensorSpec(shape=(self.n_eval_nodes_per_graph, 1), dtype=tf.float32),
                    tf.TensorSpec(shape=(self.n_eval_nodes_per_graph, self.n_node_covariates), dtype=tf.float32),
                    tf.TensorSpec(shape=(self.n_domains,), dtype=tf.int32),
                ),
                tf.TensorSpec(shape=(self.n_eval_nodes_per_graph, self.n_features_1), dtype=tf.float32),
            ),
        )
        if train:
            if shuffle_buffer_size is not None:
                dataset = dataset.shuffle(buffer_size=shuffle_buffer_size, seed=None, reshuffle_each_iteration=True)
            dataset = dataset.repeat()
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(prefetch)
        return dataset

    def _get_resampled_dataset(
        self,
        image_keys: np.ndarray,
        nodes_idx: dict,
        batch_size: int,
        seed: Optional[int] = None,
        prefetch: int = 100,
    ):
        pass

    def init_model(
        self,
        optimizer: str = "adam",
        learning_rate: float = 0.0001,
        n_eval_nodes_per_graph: int = 32,
        l2_coef: float = 0.0,
        l1_coef: float = 0.0,
        use_interactions: bool = True,
        use_domain: bool = False,
        scale_node_size: bool = False,
        output_layer: str = "linear",
        **kwargs
    ):
        self.n_eval_nodes_per_graph = n_eval_nodes_per_graph
        self.model = ModelInteractions(
            input_shapes=(
                self.n_features_0,  # target_dim
                self.n_features_1,  # out_node_feature_dim
                self.n_features_0 ** 2,  # interaction_dim
                self.n_eval_nodes_per_graph,  # in_node_dim
                self.n_node_covariates,  # categ_condition_dim
                self.n_domains,  # domain_dim
            ),
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            use_interactions=use_interactions,
            use_domain=use_domain,
            scale_node_size=scale_node_size,
            output_layer=output_layer,
        )
        optimizer = tf.keras.optimizers.get(optimizer)
        tf.keras.backend.set_value(optimizer.lr, learning_rate)
        self._compile_model(optimizer=optimizer, output_layer=output_layer)



from typing import Union

import tensorflow as tf

from ncem.models.layers import (DenseInteractions, LinearConstDispOutput,
                                LinearOutput)


class ModelInteractions:
    def __init__(
        self,
        input_shapes,
        l2_coef: Union[float, None] = 0.0,
        l1_coef: Union[float, None] = 0.0,
        use_interactions: bool = False,
        use_domain: bool = False,
        scale_node_size: bool = False,
        output_layer: str = "linear",
        **kwargs
    ):
        super().__init__()
        self.args = {
            "input_shapes": input_shapes,
            "l2_coef": l2_coef,
            "l1_coef": l1_coef,
            "use_interactions": use_interactions,
            "use_domain": use_domain,
            "scale_node_size": scale_node_size,
            "output_layer": output_layer,
        }
        target_dim = input_shapes[0]
        out_node_feature_dim = input_shapes[1]
        interaction_dim = input_shapes[2]
        in_node_dim = input_shapes[3]
        categ_condition_dim = input_shapes[4]
        domain_dim = input_shapes[5]

        input_target = tf.keras.Input(shape=(in_node_dim, target_dim), name="target")
        input_interaction = tf.keras.Input(shape=(in_node_dim, interaction_dim), name="interaction", sparse=True)
        # node size - reconstruction: Input Tensor - shape=(None, N, 1)
        input_node_size = tf.keras.Input(shape=(in_node_dim, 1), name="node_size_reconstruct")
        # Categorical predictors: Input Tensor - shape=(None, N, P)
        input_categ_condition = tf.keras.Input(shape=(in_node_dim, categ_condition_dim), name="categorical_predictor")
        # domain information of graph - shape=(None, 1)
        input_g = tf.keras.layers.Input(shape=(domain_dim,), name="input_da_group", dtype="int32")

        if use_domain:
            categ_condition = tf.concat(
                [
                    input_categ_condition,
                    tf.tile(tf.expand_dims(tf.cast(input_g, dtype="float32"), axis=-2), [1, in_node_dim, 1]),
                ],
                axis=-1,
            )
        else:
            categ_condition = input_categ_condition

        if use_interactions:
            interactions = DenseInteractions(in_node_dim=in_node_dim, interaction_dim=interaction_dim)(
                input_interaction
            )
            x = tf.concat([input_target, interactions, categ_condition], axis=-1, name="input_concat")
        else:
            x = tf.concat([input_target, categ_condition], axis=-1, name="input_concat")
        x = tf.reshape(x, [-1, x.shape[-1]], name="input_reshape")  # bs * n x (neighbour_embedding + categ_cond)

        output = tf.keras.layers.Dense(
            units=out_node_feature_dim,
            activation="linear",
            kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
            name="LinearLayer",
        )(
            x
        )  # bs * n x genes
        output = tf.reshape(output, [-1, in_node_dim, out_node_feature_dim], name="output_reshape")  # bs x n x genes

        if output_layer == "linear":
            output = LinearOutput(use_node_scale=scale_node_size, name="LinearOutput")((output, input_node_size))
        elif output_layer == "linear_const_disp":
            output = LinearConstDispOutput(use_node_scale=scale_node_size, name="LinearOutput")(
                (output, input_node_size)
            )
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)

        output_concat = tf.keras.layers.Concatenate(axis=2, name="reconstruction")(output)

        self.training_model = tf.keras.Model(
            inputs=[input_target, input_interaction, input_node_size, input_categ_condition, input_g],
            outputs=output_concat,
            name="interaction_linear_model",
        )



import math

import numpy as np
import tensorflow as tf


class NegBinLoss(tf.keras.losses.Loss):
    def call(self, y_true, y_pred):
        """Implements the negative log likelihood loss as reconstruction loss"""
        x = y_true
        loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=2)

        eta_loc = tf.math.log(loc)
        eta_scale = tf.math.log(scale)

        log_r_plus_mu = tf.math.log(scale + loc)

        ll = tf.math.lgamma(scale + x)
        ll = ll - tf.math.lgamma(x + tf.ones_like(x))
        ll = ll - tf.math.lgamma(scale)
        ll = ll + tf.multiply(x, eta_loc - log_r_plus_mu) + tf.multiply(scale, eta_scale - log_r_plus_mu)

        ll = tf.clip_by_value(ll, -300, 300, "log_probs")
        neg_ll = (tf.reduce_sum(-ll, axis=-1),)  # sum across output features
        return neg_ll


class GaussianLoss(tf.keras.losses.Loss):
    def call(self, y_true, y_pred):
        """Implements Gaussian loss as reconstruction loss"""
        y_pred, sd = tf.split(y_pred, num_or_size_splits=2, axis=2, name="gaussian_loss_split")
        # sd = 1.  # change also in metric if this is changed

        neg_ll = tf.math.log(tf.sqrt(2 * math.pi) * sd) + 0.5 * tf.math.square(y_pred - y_true) / tf.math.square(sd)
        neg_ll = tf.reduce_sum(neg_ll, axis=-1)  # sum across output features
        return neg_ll


class KLLoss(tf.keras.losses.Loss):
    def __init__(self, beta: float = 1.0, max_beta: float = 1.0, pre_warm_up: int = 0):
        super().__init__()
        self.beta = tf.Variable(beta, dtype=tf.float32, trainable=False)
        self.max_beta = max_beta
        self.pre_warm_up = pre_warm_up

    def call(self, y_true, y_pred):
        z, z_mean, z_log_var = tf.split(y_pred, num_or_size_splits=3, axis=1)
        log2pi = tf.math.log(2.0 * np.pi)
        logqz_x = -0.5 * tf.reduce_sum(tf.square(z - z_mean) * tf.exp(-z_log_var) + z_log_var + log2pi, axis=-1)
        logpz = -0.5 * tf.reduce_sum(tf.square(z) + log2pi, axis=-1)
        kl_loss = logqz_x - logpz
        return self.beta * kl_loss



import math

import tensorflow as tf
import tensorflow.experimental.numpy as tnp


def custom_mae(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    abs = tf.abs(y_true - y_pred)
    mean = tf.reduce_mean(abs)
    return mean


def custom_mse(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    se = tf.square(y_true - y_pred)
    se = tf.reduce_mean(se, axis=-1)
    return se


def custom_mean_sd(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    _, sd = tf.split(y_pred, num_or_size_splits=2, axis=2)
    sd = tf.reduce_mean(sd)
    return sd


def logp1_custom_mse(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    y_true = tf.math.log(y_true + 1.0)
    y_pred = tf.math.log(y_pred + 1.0)
    se = tf.square(y_true - y_pred)
    se = tf.reduce_mean(se, axis=-1)
    return se


def custom_mse_scaled(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, sd = tf.split(y_pred, num_or_size_splits=2, axis=2)
    se = tf.square(y_true - y_pred) / tf.square(sd)
    se = tf.reduce_mean(se, axis=-1)
    return se


def gaussian_reconstruction_loss(y_true, y_pred):
    """

    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, sd = tf.split(y_pred, num_or_size_splits=2, axis=2)
    neg_ll = tf.math.log(tf.sqrt(2 * math.pi) * sd) + 0.5 * tf.math.square(y_pred - y_true) / tf.math.square(sd)
    neg_ll = tf.reduce_sum(neg_ll, axis=-1)  # sum across output features
    return neg_ll


def nb_reconstruction_loss(y_true, y_pred):
    """Implements the negative log likelihood loss as reconstruction loss"""
    x = y_true
    loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=2)

    eta_loc = tf.math.log(loc)
    eta_scale = tf.math.log(scale)

    log_r_plus_mu = tf.math.log(scale + loc)

    ll = tf.math.lgamma(scale + x)
    ll = ll - tf.math.lgamma(x + tf.ones_like(x))
    ll = ll - tf.math.lgamma(scale)
    ll = ll + tf.multiply(x, eta_loc - log_r_plus_mu) + tf.multiply(scale, eta_scale - log_r_plus_mu)

    ll = tf.clip_by_value(ll, -300, 300, "log_probs")
    neg_ll = (tf.reduce_sum(-ll, axis=-1),)  # sum across output features
    return neg_ll


def custom_kl(y_true, y_pred):
    z, z_mean, z_log_var = tf.split(y_pred, num_or_size_splits=3, axis=1)
    log2pi = tf.math.log(2.0 * math.pi)
    logqz_x = -0.5 * tf.reduce_sum(tf.square(z - z_mean) * tf.exp(-z_log_var) + z_log_var + log2pi, axis=-1)
    logpz = -0.5 * tf.reduce_sum(tf.square(z) + log2pi, axis=-1)
    kl_loss = logqz_x - logpz
    return kl_loss


def r_squared(y_true, y_pred):
    """
    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    residual = tf.reduce_sum(tf.square(y_true - y_pred))
    total = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true)))
    r2 = tf.subtract(1.0, tf.math.divide(residual, total))
    return r2


def r_squared_linreg(y_true, y_pred):
    """
    This implementation follows the scipy linear regression implementation of R2.
    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    x = y_true
    y = y_pred
    # means
    xmean = tnp.mean(x)
    ymean = tnp.mean(y)
    # covariance
    ssxm = tnp.mean(tnp.square(x - xmean))
    ssym = tnp.mean(tnp.square(y - ymean))
    ssxym = tnp.mean((x - xmean) * (y - ymean))
    xmym = ssxm * ssym
    # Helper functions for tf.cond
    def f0():
        return tnp.zeros(shape=(), dtype=tf.float32)

    def f1():
        return tnp.ones(shape=(), dtype=tf.float32)

    def r():
        return ssxym / tnp.sqrt(xmym)  # formula for r

    def r2():
        return r ** 2  # formula for r_squared

    # R-value
    # If the denominator was going to be 0, r = 0.0
    r = tf.cond(tnp.not_equal(xmym, tnp.zeros(shape=(), dtype=tf.float32)), r, f0)
    # Test for numerical error propagation (make sure -1 < r < 1)
    r_squared = tf.cond(tnp.greater(tnp.abs(r), tnp.ones(shape=(), dtype=tf.float32)), f1, r2)
    return r_squared


def logp1_r_squared(y_true, y_pred):
    """
    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)
    y_true = tf.math.log(y_true + 1.0)
    y_pred = tf.math.log(y_pred + 1.0)
    residual = tf.reduce_sum(tf.square(y_true - y_pred))
    total = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true)))
    r2 = tf.subtract(1.0, tf.math.divide(residual, total))
    return r2


def logp1_r_squared_linreg(y_true, y_pred):
    """
    This implementation follows the scipy linear regression implementation of R2.
    :param y_true:
    :param y_pred:
    :return: (graphs,)
    """
    y_pred, _ = tf.split(y_pred, num_or_size_splits=2, axis=2)

    x = tf.math.log(y_true + 1.0)
    y = tf.math.log(y_pred + 1.0)

    # means
    xmean = tnp.mean(x)
    ymean = tnp.mean(y)

    ssxm = tnp.mean(tnp.square(x - xmean))
    ssym = tnp.mean(tnp.square(y - ymean))
    ssxym = tnp.mean((x - xmean) * (y - ymean))

    # R-value
    r = ssxym / tnp.sqrt(ssxm * ssym)
    return r ** 2



import tensorflow as tf


class LinearOutput(tf.keras.layers.Layer):
    def __init__(self, use_node_scale: bool = False, name="linear_output", **kwargs):
        super().__init__(name=name, **kwargs)
        self.use_node_scale = use_node_scale
        self.var_bias = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        genes_dim = input_shapes[0][-1]
        self.var_bias = self.add_weight("var_bias", shape=[1, genes_dim], initializer="zeros")

    def call(self, inputs, **kwargs):
        bound = 60.0
        mean, sf = inputs

        var = self.var_bias

        if self.use_node_scale:
            mean = mean * tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip")
        var = tf.zeros_like(mean) + var  # broadcast

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -tf.exp(bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        # exp_mean = mean_clip + sf
        eta_loc = mean_clip
        eta_scale = tf.exp(var_clip)

        return [eta_loc, eta_scale]


class LinearConstDispOutput(tf.keras.layers.Layer):
    def __init__(self, use_node_scale: bool = False, name="linear_const_disp_output", **kwargs):
        super().__init__(name=name, **kwargs)
        self.use_node_scale = use_node_scale
        self.var_bias = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def call(self, inputs, **kwargs):
        bound = 60.0
        mean, sf = inputs

        var = tf.zeros_like(mean)

        if self.use_node_scale:
            mean = mean * tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip")
        var = tf.zeros_like(mean) + var

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -tf.exp(bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        # exp_mean = mean_clip + sf
        eta_loc = mean_clip
        eta_scale = tf.exp(var_clip)

        return [eta_loc, eta_scale]


class GaussianOutput(tf.keras.layers.Layer):
    """Log normal likelihood output layer"""

    def __init__(self, original_dim=None, use_node_scale: bool = False, name="gaussian_output", **kwargs):

        super().__init__(name=name, **kwargs)

        self.original_dim = original_dim
        self.intermediate_dim = None
        self.use_node_scale = use_node_scale
        self.means = None
        self.var_bias = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        input_shape = input_shapes[0]
        self.intermediate_dim = input_shape[2]

        self.means = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")
        self.var_bias = self.add_weight("var_bias", shape=[1, self.original_dim], initializer="zeros")

    def call(self, inputs, **kwargs):
        bound = 60.0

        activation, sf = inputs
        in_node_dim = activation.shape[1]

        activation = tf.reshape(activation, [-1, self.intermediate_dim], name="output_layer_reshape_activation_fwdpass")

        mean = self.means(activation)
        var = self.var_bias

        mean = tf.reshape(mean, [-1, in_node_dim, self.original_dim], name="output_layer_reshape_mean")
        var = tf.zeros_like(mean) + var  # broadcast
        if self.use_node_scale:
            mean = mean * tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip")

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -tf.exp(bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        # exp_mean = mean_clip + sf
        eta_loc = mean_clip
        eta_scale = tf.exp(var_clip)

        return [eta_loc, eta_scale]


class GaussianConstDispOutput(tf.keras.layers.Layer):
    """Log normal likelihood output layer"""

    def __init__(self, original_dim=None, use_node_scale: bool = False, name="gaussian_output", **kwargs):

        super().__init__(name=name, **kwargs)

        self.original_dim = original_dim
        self.intermediate_dim = None
        self.use_node_scale = use_node_scale
        self.means = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        input_shape = input_shapes[0]
        self.intermediate_dim = input_shape[2]

        self.means = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")

    def call(self, inputs, **kwargs):
        bound = 60.0

        activation, sf = inputs
        in_node_dim = activation.shape[1]
        activation = tf.reshape(activation, [-1, self.intermediate_dim], name="output_layer_reshape_activation_fwdpass")

        mean = self.means(activation)
        mean = tf.reshape(mean, [-1, in_node_dim, self.original_dim], name="output_layer_reshape_mean")
        var = tf.zeros_like(mean)
        if self.use_node_scale:
            mean = mean * tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip")

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -tf.exp(bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        # exp_mean = mean_clip + sf
        eta_loc = mean_clip
        eta_scale = tf.exp(var_clip)

        return [eta_loc, eta_scale]


class NegBinOutput(tf.keras.layers.Layer):
    """Negative binomial output layer"""

    def __init__(self, original_dim=None, use_node_scale: bool = False, name="neg_bin_output", **kwargs):

        super().__init__(name=name, **kwargs)

        self.original_dim = original_dim
        self.intermediate_dim = None
        self.use_node_scale = use_node_scale
        self.means = None
        self.var = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        input_shape = input_shapes[0]
        self.intermediate_dim = input_shape[2]

        self.means = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")
        self.var = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")

    def call(self, inputs, **kwargs):
        bound = 60.0

        activation, sf = inputs
        activation = tf.reshape(activation, [-1, self.intermediate_dim], name="output_layer_reshape_activation_fwdpass")

        mean = self.means(activation)
        var = self.var(activation)
        if self.use_node_scale:
            mean = mean + tf.math.log(tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip"))

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        exp_mean = tf.exp(mean_clip)
        exp_var = tf.exp(var_clip)

        return [exp_mean, exp_var]


class NegBinSharedDispOutput(tf.keras.layers.Layer):
    """Negative binomial output layer with dispersion shared over features"""

    def __init__(self, original_dim=None, use_node_scale: bool = False, name="neg_bin_shared_disp_output", **kwargs):

        super().__init__(name=name, **kwargs)

        self.original_dim = original_dim
        self.intermediate_dim = None
        self.use_node_scale = use_node_scale
        self.means = None
        self.var = None
        self.var_bias = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        input_shape = input_shapes[0]
        self.intermediate_dim = input_shape[2]

        self.means = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")
        self.var_bias = self.add_weight("var_bias", shape=[1, self.original_dim], initializer="zeros")

    def call(self, inputs, **kwargs):
        bound = 60.0

        activation, sf = inputs
        in_node_dim = activation.shape[1]
        activation = tf.reshape(activation, [-1, self.intermediate_dim], name="output_layer_reshape_activation_fwdpass")

        mean = self.means(activation)
        var = self.var_bias

        mean = tf.reshape(mean, [-1, in_node_dim, self.original_dim], name="output_layer_reshape_mean")
        if self.use_node_scale:
            mean = mean + tf.math.log(tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip"))
        var = tf.zeros_like(mean) + var  # broadcast

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        exp_mean = tf.exp(mean_clip)
        exp_var = tf.exp(var_clip)

        return [exp_mean, exp_var]


class NegBinConstDispOutput(tf.keras.layers.Layer):
    """Negative binomial output layer with constant dispersion"""

    def __init__(self, original_dim=None, use_node_scale: bool = False, name="neg_bin_const_disp_output", **kwargs):

        super().__init__(name=name, **kwargs)

        self.original_dim = original_dim
        self.intermediate_dim = None
        self.use_node_scale = use_node_scale
        self.means = None
        self.var = None

    def get_config(self):
        config = super().get_config().copy()
        config.update({"original_dim": self.original_dim, "use_node_scale": self.use_node_scale})
        return config

    def build(self, input_shapes):
        input_shape = input_shapes[0]
        self.intermediate_dim = input_shape[2]

        self.means = tf.keras.layers.Dense(units=self.original_dim, use_bias=True, activation="linear")

    def call(self, inputs, **kwargs):
        bound = 60.0

        activation, sf = inputs
        in_node_dim = activation.shape[1]
        activation = tf.reshape(activation, [-1, self.intermediate_dim], name="output_layer_reshape_activation_fwdpass")

        mean = self.means(activation)
        var = tf.zeros_like(mean)

        mean = tf.reshape(mean, [-1, in_node_dim, self.original_dim], name="output_layer_reshape_mean")
        if self.use_node_scale:
            mean = mean + tf.math.log(tf.clip_by_value(sf, tf.exp(-bound), tf.exp(bound), "decoder_sf_clip"))
        var = tf.zeros_like(mean) + var  # broadcast

        # clip to log of largest values supported by log operation
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        exp_mean = tf.exp(mean_clip)
        exp_var = tf.exp(var_clip)

        return [exp_mean, exp_var]



from ncem.train.summaries import GridSearchContainer
from ncem.train.train_model import (TrainModelCVAE, TrainModelCVAEncem,
                                    TrainModelED, TrainModelEDncem,
                                    TrainModelInteractions, TrainModelLinear)



import os
import pickle
from typing import Union

from ncem.estimators import (EstimatorCVAE, EstimatorCVAEncem, EstimatorED,
                             EstimatorEDncem, EstimatorInteractions,
                             EstimatorLinear)


class TrainModel:
    estimator: Union[
        EstimatorLinear, EstimatorInteractions, EstimatorED, EstimatorEDncem, EstimatorCVAE, EstimatorCVAEncem
    ]

    @staticmethod
    def _try_save(fn: str, obj):
        with open(fn, "wb") as f:
            pickle.dump(obj=obj, file=f)

    def _save_model(self, fn: str, save_weights: bool = True):
        if save_weights:
            self.estimator.model.training_model.save_weights(fn + "_model_weights.tf")
        self._try_save(fn + "_model_args.pickle", self.estimator.model.args)

    def _save_history(self, fn: str):
        self._try_save(fn + "_history.pickle", self.estimator.history)

    def _save_hyperparam(self, fn: str):
        self._try_save(fn + "_hyperparam.pickle", self.estimator.train_hyperparam)

    def _save_predictions(self, fn: str):
        prediction = self.estimator.predict()
        self._try_save(fn + "_prediction.pickle", prediction)

    def _save_evaluation(self, fn: str):
        evaluation_val = self.estimator.evaluate_any(
            img_keys=self.estimator.img_keys_eval, node_idx=self.estimator.nodes_idx_eval
        )
        evaluation_train = self.estimator.evaluate_any(
            img_keys=self.estimator.img_keys_train, node_idx=self.estimator.nodes_idx_train
        )
        evaluation_all = self.estimator.evaluate_any(
            img_keys=self.estimator.img_keys_all, node_idx=self.estimator.nodes_idx_all
        )
        evaluations = {"train": evaluation_train, "val": evaluation_val, "all": evaluation_all}
        if len(self.estimator.img_keys_test) > 0:
            evaluation_test = self.estimator.evaluate_any(
                img_keys=self.estimator.img_keys_test, node_idx=self.estimator.nodes_idx_test
            )
            evaluations["test"] = evaluation_test
        self._try_save(fn + "_evaluation.pickle", evaluations)

    def _save_indices(self, fn: str):
        if not os.path.isfile(fn + "_indices.pickle"):
            indices = {
                "all": self.estimator.complete_img_keys,
                "test": self.estimator.img_keys_test,
                "train": self.estimator.img_keys_train,
                "val": self.estimator.img_keys_eval,
                "test_nodes": self.estimator.nodes_idx_test,
                "train_nodes": self.estimator.nodes_idx_train,
                "val_nodes": self.estimator.nodes_idx_eval,
            }
            self._try_save(fn + "_indices.pickle", indices)

    def _save_data_input(self, fn: str):
        if not os.path.isfile(fn + "_datainput.pickle"):
            info = {"img_to_patient_dict": self.estimator.img_to_patient_dict}
            self._try_save(fn + "_datainput.pickle", info)

    def save_time(self, fn: str, duration):
        with open(fn + "_time.pickle", "wb") as f:
            pickle.dump(obj=duration, file=f)

    def _save_specific(self, fn):
        # save some model specific interesting stuff (override in specific Train classes)
        pass

    def _save_evaluation_per_node_type(self, fn: str):
        split_per_node_type, evaluation_per_node_type = self.estimator.evaluate_per_node_type()
        self._try_save(fn + "_ntindices.pickle", split_per_node_type)
        self._try_save(fn + "_ntevaluation.pickle", evaluation_per_node_type)

    def save(self, fn: str, save_weights: bool = True):
        """
        Save weights and summary statistics.
        """
        assert self.estimator is not None, "initialize estimator first"
        self._save_model(fn=fn, save_weights=save_weights)

        self._save_history(fn=fn)
        self._save_hyperparam(fn=fn)
        self._save_specific(fn=fn)
        self._save_evaluation(fn=fn)

        self._save_indices(fn=fn)
        self._save_data_input(fn=fn)


class TrainModelLinear(TrainModel):
    estimator: EstimatorLinear

    def init_estim(self, **kwargs):
        self.estimator = EstimatorLinear(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_per_node_type(fn=fn)


class TrainModelInteractions(TrainModel):
    estimator: EstimatorInteractions

    def init_estim(self, **kwargs):
        self.estimator = EstimatorInteractions(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_per_node_type(fn=fn)


class TrainModelED(TrainModel):
    estimator: EstimatorED

    def init_estim(self, **kwargs):
        self.estimator = EstimatorED(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_per_node_type(fn=fn)


class TrainModelEDncem(TrainModel):
    estimator: EstimatorEDncem

    def init_estim(self, **kwargs):
        self.estimator = EstimatorEDncem(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_per_node_type(fn=fn)


class TrainModelCVAEBase(TrainModel):
    estimator: Union[EstimatorCVAE, EstimatorCVAEncem]

    def _save_evaluation_posterior_sampling(self, fn: str):
        (
            evaluation_val,
            true_val,
            pred_val,
            z_val,
            z_mean_val,
            z_log_var_val,
        ) = self.estimator.evaluate_any_posterior_sampling(
            img_keys=self.estimator.img_keys_eval, node_idx=self.estimator.nodes_idx_eval
        )
        (
            evaluation_train,
            true_train,
            pred_train,
            z_train,
            z_mean_train,
            z_log_var_train,
        ) = self.estimator.evaluate_any_posterior_sampling(
            img_keys=self.estimator.img_keys_train, node_idx=self.estimator.nodes_idx_train
        )
        (
            evaluation_all,
            true_all,
            pred_all,
            z_all,
            z_mean_all,
            z_log_var_all,
        ) = self.estimator.evaluate_any_posterior_sampling(
            img_keys=self.estimator.img_keys_all, node_idx=self.estimator.nodes_idx_all
        )
        evaluations = {"train": evaluation_train, "val": evaluation_val, "all": evaluation_all}
        predictions = {
            "train": (true_train, pred_train),
            "val": (true_val, pred_val),
            "all": (true_all, pred_all),
        }
        latent_space = {
            "train": (z_train, z_mean_train, z_log_var_train),
            "val": (z_val, z_mean_val, z_log_var_val),
            "all": (z_all, z_mean_all, z_log_var_all),
        }
        if len(self.estimator.img_keys_test) > 0:
            (
                evaluation_test,
                true_test,
                pred_test,
                z_test,
                z_mean_test,
                z_log_var_test,
            ) = self.estimator.evaluate_any_posterior_sampling(
                img_keys=self.estimator.img_keys_test, node_idx=self.estimator.nodes_idx_test
            )
            evaluations["test"] = evaluation_test
            predictions["test"] = (true_test, pred_test)
            latent_space["test"] = (z_test, z_mean_test, z_log_var_test)
        self._try_save(fn + "_evaluation_posterior_sampling.pickle", evaluations)
        self._try_save(fn + "_prediction_posterior_sampling.pickle", predictions)
        self._try_save(fn + "_latent_space_posterior_sampling.pickle", latent_space)


class TrainModelCVAE(TrainModelCVAEBase):
    estimator: EstimatorCVAE

    def init_estim(self, **kwargs):
        self.estimator = EstimatorCVAE(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_posterior_sampling(fn=fn)
        self._save_evaluation_per_node_type(fn=fn)


class TrainModelCVAEncem(TrainModelCVAEBase):
    estimator: EstimatorCVAEncem

    def init_estim(self, **kwargs):
        self.estimator = EstimatorCVAEncem(**kwargs)

    def _save_specific(self, fn):
        self._save_evaluation_posterior_sampling(fn=fn)
        self._save_evaluation_per_node_type(fn=fn)
        self._try_save(fn + "_pretrain_history.pickle", self.estimator.pretrain_history)



import numpy as np


def _get_p_value(a_var: np.ndarray, fisher_inv: np.ndarray, coef_loc_totest: int):
    """
    Returns the p-value for differential expression for each gene
    :return: np.ndarray
    """
    from diffxpy.stats.stats import wald_test

    theta_mle = a_var[coef_loc_totest]
    theta_sd = fisher_inv[:, coef_loc_totest, coef_loc_totest]
    theta_sd = np.nextafter(0, np.inf, out=theta_sd, where=theta_sd < np.nextafter(0, np.inf))
    theta_sd = np.sqrt(theta_sd)
    return wald_test(theta_mle=theta_mle, theta_sd=theta_sd, theta0=0)


def wald_test(
    parameters,
    fisher_inv: list,
    significance_threshold: float = 0.01,
):
    from diffxpy.testing.correction import correct

    res = []
    bool_res = []
    for i, fi in enumerate(fisher_inv):
        sig = []
        bool_sig = []
        a_var = parameters[:, i, :].T
        for idx in range(a_var.shape[0]):
            p_val = _get_p_value(a_var, fi, idx)
            significance = correct(p_val)
            bool_significance = significance < significance_threshold
            sig.append(significance)
            bool_sig.append(bool_significance)
        res.append(np.array(sig))  # produces (target, sources, genes) array
        bool_res.append(np.array(bool_sig))
    return np.array(bool_res), np.array(res)



from ncem.models.custom_callbacks import BetaScheduler
from ncem.models.model_cvae import ModelCVAE
from ncem.models.model_cvae_ncem import ModelCVAEncem
from ncem.models.model_ed import ModelED
from ncem.models.model_ed_ncem import ModelEDncem
from ncem.models.model_interactions import ModelInteractions
from ncem.models.model_linear import ModelLinear



from ncem.estimators import (Estimator, EstimatorCVAE, EstimatorCVAEncem,
                             EstimatorED, EstimatorEDncem, EstimatorGraph,
                             EstimatorInteractions, EstimatorLinear,
                             EstimatorNoGraph)
from ncem.models import BetaScheduler



import unittest

import ncem.api as ncem


class TestEstimator(unittest.TestCase):
    base_path = "/Users/anna.schaar/phd/datasets/"
    data_path_zhang = base_path + "zhang/"
    data_path_jarosch = base_path + "busch/"
    data_path_hartmann = base_path + "hartmann/"
    data_path_schuerch = base_path + "buffer/"

    def get_estimator(
        self,
        model: str,
        data_origin: str = "zhang",
    ):
        node_label_space_id = "type"
        node_feature_space_id = "standard"
        if model in ["linear_baseline", "linear"]:
            self.est = ncem.train.EstimatorLinear()
        elif model in ["interactions_baseline", "interactions"]:
            self.est = ncem.train.EstimatorInteractions()
        elif model == "ed":
            self.est = ncem.train.EstimatorED()
        elif model == "ed_ncem_max":
            self.est = ncem.train.EstimatorEDncem(cond_type="max")
        elif model == "ed_ncem_gcn":
            self.est = ncem.train.EstimatorEDncem(cond_type="gcn")
        elif model == "cvae":
            self.est = ncem.train.EstimatorCVAE()
        elif model == "cvae_ncem_max":
            self.est = ncem.train.EstimatorCVAEncem(cond_type="max")
        elif model == "cvae_ncem_gcn":
            self.est = ncem.train.EstimatorCVAEncem(cond_type="gcn")
        else:
            assert False

        if data_origin == "zhang":
            radius = 100
            data_path = self.data_path_zhang
        elif data_origin == "hartmann":
            radius = 100
            data_path = self.data_path_hartmann
        else:
            assert False

        self.est.get_data(
            data_origin=data_origin,
            data_path=data_path,
            radius=radius,
            node_label_space_id=node_label_space_id,
            node_feature_space_id=node_feature_space_id,
            merge_node_types_predefined=True,
        )

    def _test_train(self, model: str, data_origin: str = "zhang"):
        self.get_estimator(model=model, data_origin=data_origin)

        if model == "linear":
            kwargs = {"use_source_type": True, "use_domain": True, "learning_rate": 1e-2}
            train_kwargs = {}
        elif model == "linear_baseline":
            kwargs = {"use_source_type": False, "use_domain": True, "learning_rate": 1e-2}
            train_kwargs = {}
        elif model == "interactions":
            kwargs = {"use_interactions": True, "use_domain": True, "learning_rate": 1e-2}
            train_kwargs = {}
        elif model == "interactions_baseline":
            kwargs = {"use_interactions": False, "use_domain": True, "learning_rate": 1e-2}
            train_kwargs = {}
        elif model == "ed":
            kwargs = {
                "depth_enc": 1,
                "depth_dec": 0,
                "use_domain": True,
                "use_bias": True,
                "learning_rate": 1e-2,
                "beta": 0.1,
            }
            train_kwargs = {}
        elif model in ["ed_ncem_max", "ed_ncem_gcn"]:
            kwargs = {
                "depth_enc": 1,
                "depth_dec": 0,
                "cond_depth": 1,
                "use_domain": True,
                "use_bias": True,
                "learning_rate": 1e-2,
                "beta": 0.1,
            }
            train_kwargs = {}
        elif model == "cvae":
            kwargs = {
                "depth_enc": 1,
                "depth_dec": 1,
                "use_domain": True,
                "use_bias": True,
                "learning_rate": 1e-2,
                "beta": 0.1,
            }
            train_kwargs = {}
        elif model in ["cvae_ncem_max", "cvae_ncem_gcn"]:
            kwargs = {
                "depth_enc": 1,
                "depth_dec": 1,
                "cond_depth": 1,
                "use_domain": True,
                "use_bias": True,
                "learning_rate": 1e-2,
                "beta": 0.1,
            }
            train_kwargs = {}
        else:
            assert False
        self._model_kwargs = kwargs
        self.est.init_model(**kwargs)
        self.est.split_data_node(validation_split=0.5, test_split=0.5)

        if data_origin == "hartmann":
            batch_size = None
        else:
            batch_size = 16

        if batch_size is None:
            bs = len(list(self.est.complete_img_keys))
            shuffle_buffer_size = None  # None
        else:
            bs = batch_size
            shuffle_buffer_size = int(2)
        self.est.train(
            epochs=5,
            max_steps_per_epoch=2,
            batch_size=bs,
            validation_batch_size=6,
            max_validation_steps=1,
            shuffle_buffer_size=shuffle_buffer_size,
            log_dir=None,
            **train_kwargs
        )
        self.est.model.training_model.summary()

    def test_linear(self):
        # self._test_train(model='linear_baseline', data_origin='hartmann')
        # self._test_train(model="linear", data_origin="hartmann")
        # self._test_train(model='interactions_baseline', data_origin='hartmann')
        self._test_train(model="interactions", data_origin="hartmann")

    def test_ed(self):
        # self._test_train(model="ed", data_origin="hartmann")
        self._test_train(model="ed_ncem_max", data_origin="hartmann")
        # self._test_train(model="ed_ncem_gcn", data_origin="hartmann")

    def test_cvae(self):
        # self._test_train(model="cvae", data_origin="hartmann")
        self._test_train(model="cvae_ncem_max", data_origin="hartmann")
        # self._test_train(model="cvae_ncem_gcn", data_origin="hartmann")
